#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2018-2021

# from pilot.util.jobmetrics import get_job_metrics_entry

import logging
logger = logging.getLogger(__name__)


def get_job_metrics(job):
    """
    Return a properly formatted job metrics string.
    The format of the job metrics string is defined by the server. It will be reported to the server during updateJob.

    Example of job metrics:
    Number of events read | Number of events written | vmPeak maximum | vmPeak average | RSS average | ..
    Format: nEvents=<int> nEventsW=<int> vmPeakMax=<int> vmPeakMean=<int> RSSMean=<int> hs06=<float> shutdownTime=<int>
            cpuFactor=<float> cpuLimit=<float> diskLimit=<float> jobStart=<int> memLimit=<int> runLimit=<float>

    :param job: job object.
    :return: job metrics (string).
    """

    return ""



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2018-2022

import os
import time
import signal
import re
import threading

from pilot.util.container import execute
from pilot.util.auxiliary import whoami
from pilot.util.filehandling import read_file, remove_dir_tree

import logging
logger = logging.getLogger(__name__)


def find_processes_in_group(cpids, pid):
    """
    Find all processes that belong to the same group.
    Recursively search for the children processes belonging to pid and return their pid's.
    pid is the parent pid and cpids is a list that has to be initialized before calling this function and it contains
    the pids of the children AND the parent.

    :param cpids: list of pid's for all child processes to the parent pid, as well as the parent pid itself (int).
    :param pid: parent process id (int).
    :return: (updated cpids input parameter list).
    """

    if not pid:
        return

    cpids.append(pid)

    cmd = "ps -eo pid,ppid -m | grep %d" % pid
    exit_code, psout, stderr = execute(cmd, mute=True)

    lines = psout.split("\n")
    if lines != ['']:
        for i in range(0, len(lines)):
            try:
                thispid = int(lines[i].split()[0])
                thisppid = int(lines[i].split()[1])
            except Exception as error:
                logger.warning(f'exception caught: {error}')
            if thisppid == pid:
                find_processes_in_group(cpids, thispid)


def is_zombie(pid):
    """
    Is the given process a zombie?
    :param pid: process id (int).
    :return: boolean.
    """

    status = False

    cmd = "ps aux | grep %d" % (pid)
    exit_code, stdout, stderr = execute(cmd, mute=True)
    if "<defunct>" in stdout:
        status = True

    return status


def get_process_commands(euid, pids):
    """
    Return a list of process commands corresponding to a pid list for user euid.

    :param euid: user id (int).
    :param pids: list of process id's.
    :return: list of process commands.
    """

    cmd = 'ps u -u %d' % euid
    process_commands = []
    exit_code, stdout, stderr = execute(cmd, mute=True)

    if exit_code != 0 or stdout == '':
        logger.warning('ps command failed: %d, \"%s\", \"%s\"', exit_code, stdout, stderr)
    else:
        # extract the relevant processes
        p_commands = stdout.split('\n')
        first = True
        for p_command in p_commands:
            if first:
                # get the header info line
                process_commands.append(p_command)
                first = False
            else:
                # remove extra spaces
                _p_command = p_command
                while "  " in _p_command:
                    _p_command = _p_command.replace("  ", " ")
                items = _p_command.split(" ")
                for pid in pids:
                    # items = username pid ...
                    if items[1] == str(pid):
                        process_commands.append(p_command)
                        break

    return process_commands


def dump_stack_trace(pid):
    """
    Execute the stack trace command (pstack <pid>).

    :param pid: process id (int).
    :return:
    """

    # make sure that the process is not in a zombie state
    if not is_zombie(pid):
        cmd = "pstack %d" % (pid)
        exit_code, stdout, stderr = execute(cmd, mute=True, timeout=60)
        logger.info(stdout or "(pstack returned empty string)")
    else:
        logger.info("skipping pstack dump for zombie process")


def kill_processes(pid):
    """
    Kill process beloging to given process group.

    :param pid: process id (int).
    :return:
    """

    # if there is a known subprocess pgrp, then it should be enough to kill the group in one go
    status = False
    try:
        pgrp = os.getpgid(pid)
    except Exception:
        pgrp = 0
    if pgrp != 0:
        status = kill_process_group(pgrp)

    if not status:
        # firstly find all the children process IDs to be killed
        children = []
        find_processes_in_group(children, pid)

        # reverse the process order so that the athena process is killed first (otherwise the stdout will be truncated)
        if not children:
            return

        children.reverse()
        logger.info("process IDs to be killed: %s (in reverse order)", str(children))

        # find which commands are still running
        try:
            cmds = get_process_commands(os.geteuid(), children)
        except Exception as error:
            logger.warning("get_process_commands() threw an exception: %s", error)
        else:
            if len(cmds) <= 1:
                logger.warning("found no corresponding commands to process id(s)")
            else:
                logger.info("found commands still running:")
                for cmd in cmds:
                    logger.info(cmd)

                # loop over all child processes
                for i in children:
                    # dump the stack trace before killing it
                    dump_stack_trace(i)

                    # kill the process gracefully
                    kill_process(i)

    # kill any remaining orphan processes
    # note: this should no longer be necessary since ctypes has made sure all subprocesses are parented
    # if orphan process killing is not desired, set env var PILOT_NOKILL
    kill_orphans()


def kill_child_processes(pid):
    """
    Kill child processes.

    :param pid: process id (int).
    :return:
    """
    # firstly find all the children process IDs to be killed
    children = []
    find_processes_in_group(children, pid)

    # reverse the process order so that the athena process is killed first (otherwise the stdout will be truncated)
    children.reverse()
    logger.info("process IDs to be killed: %s (in reverse order)", str(children))

    # find which commands are still running
    try:
        cmds = get_process_commands(os.geteuid(), children)
    except Exception as error:
        logger.warning("get_process_commands() threw an exception: %s", error)
    else:
        if len(cmds) <= 1:
            logger.warning("found no corresponding commands to process id(s)")
        else:
            logger.info("found commands still running:")
            for cmd in cmds:
                logger.info(cmd)

            # loop over all child processes
            for i in children:
                # dump the stack trace before killing it
                dump_stack_trace(i)

                # kill the process gracefully
                kill_process(i)


def kill_process_group(pgrp):
    """
    Kill the process group.

    :param pgrp: process group id (int).
    :return: boolean (True if SIGMTERM followed by SIGKILL signalling was successful)
    """

    status = False
    _sleep = True

    # kill the process gracefully
    logger.info("killing group process %d", pgrp)
    try:
        os.killpg(pgrp, signal.SIGTERM)
    except Exception as error:
        logger.warning("exception thrown when killing child group process under SIGTERM: %s", error)
        _sleep = False
    else:
        logger.info("SIGTERM sent to process group %d", pgrp)

    if _sleep:
        _t = 30
        logger.info("sleeping %d s to allow processes to exit", _t)
        time.sleep(_t)

    try:
        os.killpg(pgrp, signal.SIGKILL)
    except Exception as error:
        logger.warning("exception thrown when killing child group process with SIGKILL: %s", error)
    else:
        logger.info("SIGKILL sent to process group %d", pgrp)
        status = True

    return status


def kill_process(pid):
    """
    Kill process.

    :param pid: process id (int).
    :return: boolean (True if successful SIGKILL)
    """

    status = False

    # start with soft kill (ignore any returned status)
    kill(pid, signal.SIGTERM)

    _t = 10
    logger.info("sleeping %d s to allow process to exit", _t)
    time.sleep(_t)

    # now do a hard kill just in case some processes haven't gone away
    status = kill(pid, signal.SIGKILL)

    return status


def kill(pid, sig):
    """
    Kill the given process with the given signal.

    :param pid: process id (int).
    :param sig: signal (int).
    :return status: True when successful (Boolean).
    """

    status = False
    try:
        os.kill(pid, sig)
    except Exception as error:
        logger.warning("exception thrown when killing process %d with signal=%d: %s", pid, sig, error)
    else:
        logger.info("killed process %d with signal=%d", pid, sig)
        status = True

    return status


# called checkProcesses() in Pilot 1, used by process monitoring
def get_number_of_child_processes(pid):
    """
    Get the number of child processes for a given parent process.

    :param pid: parent process id (int).
    :return: number of child processes (int).
    """

    children = []
    n = 0
    try:
        find_processes_in_group(children, pid)
    except Exception as error:
        logger.warning("exception caught in find_processes_in_group: %s", error)
    else:
        if pid:
            n = len(children)
            logger.info("number of running child processes to parent process %d: %d", pid, n)
        else:
            logger.debug("pid not yet set")
    return n


def killpg(pid, sig, args):
    """
    Kill given process group with given signal.

    :param pid: process group id (int).
    :param sig: signal (int).
    :return:
    """

    try:
        os.killpg(int(pid), sig)
    except Exception as error:
        logger.warning("failed to execute killpg(): %s", error)
        cmd = 'kill -%d %s' % (sig, pid)
        exit_code, rs, stderr = execute(cmd)
        if exit_code != 0:
            logger.warning(rs)
        else:
            logger.info("killed orphaned process %s (%s)", pid, args)
    else:
        logger.info("killed orphaned process group %s (%s)", pid, args)


def get_pilot_pid_from_processes(_processes, pattern):
    """
    Identify the pilot pid from the list of processes.

    :param _processes: ps output (string).
    :param pattern: regex pattern (compiled regex string).
    :return: pilot pid (int or None).
    """

    pilot_pid = None
    for line in _processes.split('\n'):
        ids = pattern.search(line)
        if ids:
            pid = ids.group(1)
            args = ids.group(3)
            try:
                pid = int(pid)
            except Exception as error:
                logger.warning('failed to convert pid to int: %s', error)
                continue
            if 'pilot.py' in args and 'python' in args:
                pilot_pid = pid
                break

    return pilot_pid


def kill_orphans():
    """
    Find and kill all orphan processes belonging to current pilot user.

    :return:
    """

    # exception for BOINC
    if 'BOINC' in os.environ.get('PILOT_SITENAME', ''):
        logger.info("Do not look for orphan processes in BOINC jobs")
        return

    if 'PILOT_NOKILL' in os.environ:
        return

    logger.info("searching for orphan processes")

    cmd = "ps -o pid,ppid,args -u %s" % whoami()
    exit_code, _processes, stderr = execute(cmd)
    pattern = re.compile(r'(\d+)\s+(\d+)\s+([\S\s]+)')

    count = 0
    for line in _processes.split('\n'):
        ids = pattern.search(line)
        if ids:
            pid = ids.group(1)
            ppid = ids.group(2)
            args = ids.group(3)
            try:
                pid = int(pid)
            except Exception as error:
                logger.warning('failed to convert pid to int: %s', error)
                continue
            if 'cvmfs2' in args:
                logger.info("ignoring possible orphan process running cvmfs2: pid=%s, ppid=%s, args=\'%s\'", pid, ppid, args)
            elif 'pilots_starter.py' in args or 'runpilot2-wrapper.sh' in args or 'runpilot3-wrapper.sh' in args:
                logger.info("ignoring pilot launcher: pid=%s, ppid=%s, args='%s'", pid, ppid, args)
            elif ppid == '1':
                count += 1
                logger.info("found orphan process: pid=%s, ppid=%s, args='%s'", pid, ppid, args)
                if 'bash' in args or ('python' in args and 'pilot.py' in args):
                    logger.info("will not kill bash process")
                else:
                    killpg(pid, signal.SIGTERM, args)
                    _t = 10
                    logger.info("sleeping %d s to allow processes to exit", _t)
                    time.sleep(_t)
                    killpg(pid, signal.SIGKILL, args)

    if count == 0:
        logger.info("did not find any orphan processes")
    else:
        logger.info("found %d orphan process(es)", count)


def get_max_memory_usage_from_cgroups():
    """
    Read the max_memory from CGROUPS file memory.max_usage_in_bytes.

    :return: max_memory (int).
    """

    max_memory = None

    # Get the CGroups max memory using the pilot pid
    pid = os.getpid()
    path = "/proc/%d/cgroup" % pid
    if os.path.exists(path):
        cmd = "grep memory %s" % path
        exit_code, out, stderr = execute(cmd)
        if out == "":
            logger.info("(command did not return anything)")
        else:
            logger.info(out)
            if ":memory:" in out:
                pos = out.find('/')
                path = out[pos:]
                logger.info("extracted path = %s", path)

                pre = get_cgroups_base_path()
                if pre != "":
                    path = pre + os.path.join(path, "memory.max_usage_in_bytes")
                    logger.info("path to CGROUPS memory info: %s", path)
                    max_memory = read_file(path)
                else:
                    logger.info("CGROUPS base path could not be extracted - not a CGROUPS site")
            else:
                logger.warning("invalid format: %s (expected ..:memory:[path])", out)
    else:
        logger.info("path %s does not exist (not a CGROUPS site)", path)

    return max_memory


def get_cgroups_base_path():
    """
    Return the base path for CGROUPS.

    :return: base path for CGROUPS (string).
    """

    cmd = "grep \'^cgroup\' /proc/mounts|grep memory| awk \'{print $2}\'"
    exit_code, base_path, stderr = execute(cmd, mute=True)

    return base_path


def get_cpu_consumption_time(t0):
    """
    Return the CPU consumption time for child processes measured by system+user time from os.times().
    Note: the os.times() tuple is user time, system time, s user time, s system time, and elapsed real time since a
    fixed point in the past.

    :param t0: initial os.times() tuple prior to measurement.
    :return: system+user time for child processes (float).
    """

    t1 = os.times()
    user_time = t1[2] - t0[2]
    system_time = t1[3] - t0[3]

    return user_time + system_time


def get_instant_cpu_consumption_time(pid):
    """
    Return the CPU consumption time (system+user time) for a given process, by parsing /prod/pid/stat.
    Note 1: the function returns 0.0 if the pid is not set.
    Note 2: the function must sum up all the user+system times for both the main process (pid) and the child
    processes, since the main process is most likely spawning new processes.

    :param pid: process id (int).
    :return: system+user time for a given pid (float).
    """

    utime = None
    stime = None
    cutime = None
    cstime = None

    hz = os.sysconf(os.sysconf_names['SC_CLK_TCK'])
    if type(hz) != int:
        logger.warning('unknown SC_CLK_TCK: %s', str(hz))
        return 0.0

    if pid and hz and hz > 0:
        path = "/proc/%d/stat" % pid
        if os.path.exists(path):
            try:
                with open(path) as fp:
                    fields = fp.read().split(' ')[13:17]
                    utime, stime, cutime, cstime = [(float(f) / hz) for f in fields]
            except (FileNotFoundError, IOError) as exc:
                logger.warning(f'exception caught: {exc} (ignored)')

    if utime and stime and cutime and cstime:
        # sum up all the user+system times for both the main process (pid) and the child processes
        cpu_consumption_time = utime + stime + cutime + cstime
    else:
        cpu_consumption_time = 0.0

    return cpu_consumption_time


def get_current_cpu_consumption_time(pid):
    """
    Get the current CPU consumption time (system+user time) for a given process, by looping over all child processes.

    :param pid: process id (int).
    :return: system+user time for a given pid (float).
    """

    # get all the child processes
    children = []
    find_processes_in_group(children, pid)

    cpuconsumptiontime = 0
    for _pid in children:
        _cpuconsumptiontime = get_instant_cpu_consumption_time(_pid)
        if _cpuconsumptiontime:
            cpuconsumptiontime += _cpuconsumptiontime

    return cpuconsumptiontime


def is_process_running(process_id):
    """
    Check whether process is still running.

    :param process_id: process id (int).
    :return: Boolean.
    """
    try:
        # note that this kill function call will not kill the process
        os.kill(process_id, 0)
        return True
    except OSError:
        return False


def cleanup(job, args):
    """
    Cleanup called after completion of job.

    :param job: job object
    :return:
    """

    logger.info("overall cleanup function is called")

    # make sure the workdir is deleted
    if args.cleanup:
        if remove_dir_tree(job.workdir):
            logger.info('removed %s', job.workdir)

        if os.path.exists(job.workdir):
            logger.warning('work directory still exists: %s', job.workdir)
        else:
            logger.debug('work directory was removed: %s', job.workdir)
    else:
        logger.info('workdir not removed %s', job.workdir)

    # collect any zombie processes
    job.collect_zombies(depth=10)
    logger.info("collected zombie processes")

    if job.pid:
        logger.info("will now attempt to kill all subprocesses of pid=%d", job.pid)
        kill_processes(job.pid)
    else:
        logger.warning('cannot kill any subprocesses since job.pid is not set')
    #logger.info("deleting job object")
    #del job


def threads_aborted_deprecated(abort_at=2):
    """
    Have the threads been aborted?

    :param abort_at: 1 for workflow finish, 2 for thread finish (since check is done just before thread finishes) (int).
    :return: Boolean.
    """

    aborted = False
    thread_count = threading.activeCount()

    # count all non-daemon threads
    daemon_threads = 0
    for thread in threading.enumerate():
        _thr = ''
        if thread.isDaemon():  # ignore any daemon threads, they will be aborted when python ends
            if abort_at == 1:
                _thr = f'thread={thread} (daemon)'
            daemon_threads += 1
        else:
            if abort_at == 1:
                _thr = f'thread={thread}'
        if _thr:
            _thr += f' (thread_count={thread_count}, daemon_threads={daemon_threads}, abort_at={abort_at})'
            logger.debug(_thr)
    if thread_count - daemon_threads == abort_at:
        logger.debug(f'aborting since the last relevant thread is about to finish ({thread_count} - {daemon_threads} = {abort_at})')
        aborted = True

    return aborted


def threads_aborted():
    """
    Have the Pilot threads been aborted?
    This function will count all the threads still running, but will only return True if all
    threads started by the Pilot's main thread, i.e. not including the main thread itself or
    any daemon threads (which might be created by Rucio or Google Logging).

    :return: True if number of running threads is zero (Boolean).
    """

    abort = False
    thread_count = threading.activeCount()
    pilot_thread_count = 0
    daemon_threads = 0

    # count all threads still alive
    for thread in threading.enumerate():
        if thread.isDaemon():  # ignore any daemon threads, they will be aborted when python ends
            daemon_threads += 1
            #tag = 'daemon'
        elif thread == threading.main_thread():
            #tag = 'main'
            pass
        else:  # only count threads spawned by the main thread, no the main thread itself or any daemon threads
            #tag = 'pilot?'
            pilot_thread_count += 1
        #logger.debug(f'thread={thread}, pilot_thread_count={pilot_thread_count}, daemon_thread_count={daemon_threads}, tag={tag}')
    if pilot_thread_count == 0:
        logger.debug(f'aborting since only the main Pilot thread is still running '
                     f'(total thread count={thread_count} with {daemon_threads} daemon thread(s)')
        abort = True

    return abort


def convert_ps_to_dict(output, pattern=r'(\d+) (\d+) (\d+) (.+)'):
    """
    Convert output from a ps command to a dictionary.

    Example: ps axo pid,ppid,pgid,cmd
      PID  PPID  PGID COMMAND
      22091  6672 22091 bash
      32581 22091 32581 ps something;sdfsdfds/athena.py ddfg
      -> dictionary = { 'PID': [22091, 32581], 'PPID': [22091, 6672], .. , 'COMMAND': ['ps ..', 'bash']}

    :param output: ps stdout (string).
    :param pattern: regex pattern matching the ps output (raw string).
    :return: dictionary.
    """

    dictionary = {}
    first_line = []  # e.g. PID PPID PGID COMMAND

    for line in output.split('\n'):
        try:
            # remove leading and trailing spaces
            line = line.strip()
            # remove multiple spaces inside the line
            _l = re.sub(' +', ' ', line)

            if first_line == []:
                _l = [_f for _f in _l.split(' ') if _f]
                first_line = _l
                for i in range(len(_l)):
                    dictionary[_l[i]] = []
            else:  # e.g. 22091 6672 22091 bash
                match = re.search(pattern, _l)
                if match:
                    for i in range(len(first_line)):
                        try:
                            var = int(match.group(i + 1))
                        except Exception:
                            var = match.group(i + 1)
                        dictionary[first_line[i]].append(var)

        except Exception as error:
            print("unexpected format of utility output: %s", error)

    return dictionary


def get_trimmed_dictionary(keys, dictionary):
    """
    Return a sub-dictionary with only the given keys.

    :param keys: keys to keep (list).
    :param dictionary: full dictionary.
    :return: trimmed dictionary.
    """

    subdictionary = {}
    for key in keys:
        if key in dictionary:
            subdictionary[key] = dictionary[key]

    return subdictionary


def find_cmd_pids(cmd, ps_dictionary):
    """
    Find all pids for the given command.
    Example. cmd = 'athena.py' -> pids = [1234, 2267] (in case there are two pilots running on the WN).

    :param cmd: command (string).
    :param ps_dictionary: converted ps output (dictionary).
    """

    pids = []
    i = -1
    for _cmd in ps_dictionary.get('COMMAND'):
        i += 1
        if cmd in _cmd:
            pids.append(ps_dictionary.get('PID')[i])
    return pids


def find_pid(pandaid, ps_dictionary):
    """
    Find the process id for the command that contains 'export PandaID=%d'.

    :param pandaid: PanDA ID (string).
    :param ps_dictionaryL ps output dictionary.
    :return: pid (int).
    """

    pid = -1
    i = -1
    pandaid_cmd = 'export PandaID=%s' % pandaid
    for _cmd in ps_dictionary.get('COMMAND'):
        i += 1
        if pandaid_cmd in _cmd:
            pid = ps_dictionary.get('PID')[i]
            break

    return pid


def is_child(pid, pandaid_pid, dictionary):
    """
    Is the given pid a child process of the pandaid_pid?
    Proceed recursively until the parent pandaid_pid has been found, or return False if it fails to find it.
    """

    try:
        # where are we at in the PID list?
        index = dictionary.get('PID').index(pid)
    except ValueError:
        # not in the list
        return False
    else:
        # get the corresponding ppid
        ppid = dictionary.get('PPID')[index]

        print(index, pid, ppid, pandaid_pid)
        # is the current parent the same as the pandaid_pid? if yes, we are done
        if ppid == pandaid_pid:
            return True
        else:
            # try another pid
            return is_child(ppid, pandaid_pid, dictionary)


def get_subprocesses(pid):
    """
    Return the subprocesses belonging to the given PID.

    :param pid: main process PID (int).
    :return: list of subprocess PIDs.
    """

    cmd = f'ps -opid --no-headers --ppid {pid}'
    _, out, _ = execute(cmd)
    return [int(line) for line in out.splitlines()] if out else []



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Mario Lassnig, mario.lassnig@cern.ch, 2016-2017
# - Tobias Wegner, tobias.wegner@cern.ch, 2017
# - Paul Nilsson, paul.nilsson@cern.ch, 2021

import os
import random
import shutil
import tempfile
import unittest
import uuid

from pilot.api import data


def check_env():
    """
    Function to check whether cvmfs is available.
    To be used to decide whether to skip some test functions.

    :returns True: if unit test should run (currently broken)
    """
    return False


@unittest.skipIf(not check_env(), "This unit test is broken")
class TestHarvesterStageIn(unittest.TestCase):
    """
    Automatic stage-in tests for Harvester.

        from pilot.api import data
        data_client = data.StageInClient(site)
        result = data_client.transfer(files=[{scope, name, destination}, ...])

    Notabene:
      The following datasets with their constituent files are replicated
      on every DATADISK and should thus be generally available:

        user.mlassnig:user.mlassnig.pilot.test.single.hits
          mc15_13TeV:HITS.06828093._000096.pool.root.1

        user.mlassnig:user.mlassnig.pilot.test.multi.hits
          mc15_14TeV:HITS.10075481._000432.pool.root.1
          mc15_14TeV:HITS.10075481._000433.pool.root.1
          mc15_14TeV:HITS.10075481._000434.pool.root.1
          mc15_14TeV:HITS.10075481._000435.pool.root.1
          mc15_14TeV:HITS.10075481._000444.pool.root.1
          mc15_14TeV:HITS.10075481._000445.pool.root.1
          mc15_14TeV:HITS.10075481._000451.pool.root.1
          mc15_14TeV:HITS.10075481._000454.pool.root.1
          mc15_14TeV:HITS.10075481._000455.pool.root.1
    """

    def setUp(self):
        # skip tests if running through Travis -- github does not have working rucio
        self.travis = os.environ.get('TRAVIS') == 'true'

        # setup pilot data client

        # 1st example: using StageIn client with infosys
        # initialize StageInClient using infosys component to resolve allowed input sources
        #from pilot.info import infosys
        #infosys.init('ANALY_CERN')
        #self.data_client = data.StageInClient(infosys)

        # 2nd example: avoid using infosys instance but it requires to pass explicitly copytools and allowed input Strorages in order to resolve replicas
        #self.data_client = data.StageInClient(acopytools={'pr':'rucio'})
        self.data_client = data.StageInClient(acopytools='rucio')  ## use rucio everywhere

    def test_stagein_sync_fail_nodirectory(self):
        '''
        Test error message propagation.
        '''
        if self.travis:
            return True

        result = self.data_client.transfer(files=[{'scope': 'does_not_matter',
                                                   'name': 'does_not_matter',
                                                   'destination': '/i_do_not_exist'},
                                                  {'scope': 'does_not_matter_too',
                                                   'name': 'does_not_matter_too',
                                                   'destination': '/neither_do_i'}])

        self.assertIsNotNone(result)
        for _file in result:
            self.assertEqual(_file['errno'], 1)
            self.assertEqual(_file['status'], 'failed')
            #self.assertIn(_file['errmsg'], ['Destination directory does not exist: /i_do_not_exist',
            #                              'Destination directory does not exist: /neither_do_i'])

    def test_stagein_sync_fail_noexist(self):
        '''
        Test error message propagation.
        '''
        if self.travis:
            return True

        result = self.data_client.transfer(files=[{'scope': 'no_scope1',
                                                   'name': 'no_name1',
                                                   'destination': '/tmp'},
                                                  {'scope': 'no_scope2',
                                                   'name': 'no_name2',
                                                   'destination': '/tmp'}])

        self.assertIsNotNone(result)
        for _file in result:
            self.assertEqual(_file['errno'], 3)
            self.assertEqual(_file['status'], 'failed')
            #self.assertIn(_file['errmsg'], ['Data identifier \'no_scope1:no_name1\' not found',
            #                               'Data identifier \'no_scope2:no_name2\' not found'])

    def test_stagein_sync_fail_mix(self):
        '''
        Test error message propagation
        '''
        if self.travis:
            return True

        ## if infosys was not passed to StageInClient in constructor
        ## then it's mandatory to specify allowed `inputddms` that can be used as source for replica lookup
        tmp_dir1, tmp_dir2 = tempfile.mkdtemp(), tempfile.mkdtemp()
        result = self.data_client.transfer(files=[{'scope': 'no_scope1',
                                                   'name': 'no_name1',
                                                   'destination': '/tmp'},
                                                  {'scope': 'mc15_13TeV',
                                                   'name': 'HITS.06828093._000096.pool.root.1',
                                                   'destination': tmp_dir1},
                                                  {'scope': 'mc15_13TeV',
                                                   'name': 'HITS.06828093._000096.pool.root.1',
                                                   'destination': tmp_dir2},
                                                  {'scope': 'no_scope2',
                                                   'name': 'no_name2',
                                                   'destination': '/tmp'}])
        ls_tmp_dir1 = os.listdir(tmp_dir1)
        ls_tmp_dir2 = os.listdir(tmp_dir2)
        shutil.rmtree(tmp_dir1)
        shutil.rmtree(tmp_dir2)
        self.assertIn('HITS.06828093._000096.pool.root.1', ls_tmp_dir1)
        self.assertIn('HITS.06828093._000096.pool.root.1', ls_tmp_dir2)

        self.assertIsNotNone(result)
        for _file in result:
            if _file['name'] in ['no_name1', 'no_name2']:
                self.assertEqual(_file['errno'], 3)
                self.assertEqual(_file['status'], 'failed')
                #self.assertIn(_file['errmsg'], ['Data identifier \'no_scope1:no_name1\' not found',
                #                               'Data identifier \'no_scope2:no_name2\' not found'])
            else:
                self.assertEqual(_file['errno'], 0)
                self.assertEqual(_file['status'], 'done')

    def test_stagein_sync_simple(self):
        '''
        Single file going to a destination directory.
        '''
        if self.travis:
            return True

        result = self.data_client.transfer(files=[{'scope': 'mc15_13TeV',
                                                   'name': 'HITS.06828093._000096.pool.root.1',
                                                   'destination': '/tmp'}])

        os.remove('/tmp/HITS.06828093._000096.pool.root.1')

        self.assertIsNotNone(result)
        for _file in result:
            self.assertEqual(_file['errno'], 0)

    def test_stagein_sync_merged_same(self):
        '''
        Multiple files going to the same destination directory.
        '''
        if self.travis:
            return True

        result = self.data_client.transfer(files=[{'scope': 'mc15_14TeV',
                                                   'name': 'HITS.10075481._000432.pool.root.1',
                                                   'destination': '/tmp'},
                                                  {'scope': 'mc15_14TeV',
                                                   'name': 'HITS.10075481._000433.pool.root.1',
                                                   'destination': '/tmp'}])

        os.remove('/tmp/HITS.10075481._000432.pool.root.1')
        os.remove('/tmp/HITS.10075481._000433.pool.root.1')

        self.assertIsNotNone(result)
        for _file in result:
            self.assertEqual(_file['errno'], 0)

    def test_stagein_sync_merged_diff(self):
        '''
        Multiple files going to different destination directories.
        '''
        if self.travis:
            return True

        tmp_dir1, tmp_dir2 = tempfile.mkdtemp(), tempfile.mkdtemp()
        result = self.data_client.transfer(files=[{'scope': 'mc15_14TeV',
                                                   'name': 'HITS.10075481._000432.pool.root.1',
                                                   'destination': tmp_dir1},
                                                  {'scope': 'mc15_14TeV',
                                                   'name': 'HITS.10075481._000433.pool.root.1',
                                                   'destination': tmp_dir2}])

        ls_tmp_dir1 = os.listdir(tmp_dir1)
        ls_tmp_dir2 = os.listdir(tmp_dir2)
        shutil.rmtree(tmp_dir1)
        shutil.rmtree(tmp_dir2)

        self.assertIsNotNone(result)
        for _file in result:
            self.assertEqual(_file['errno'], 0)
            self.assertIn('HITS.10075481._000432.pool.root.1', ls_tmp_dir1)
            self.assertIn('HITS.10075481._000433.pool.root.1', ls_tmp_dir2)


@unittest.skipIf(not check_env(), "This unit test is broken")
class TestHarvesterStageOut(unittest.TestCase):
    '''
    Automatic stage-out tests for Harvester.

        from pilot.api import data
        data_client = data.StageOutClient(site)
        result = data_client.transfer(files=[{scope, name, ...}, ...])
    '''

    def setUp(self):
        # skip tests if running through Travis -- github does not have working rucio

        self.travis = os.environ.get('TRAVIS') == 'true'

        # setup pilot data client
        self.data_client = data.StageOutClient(acopytools=['rucio'])

    def test_stageout_fail_notfound(self):
        '''
        Test error message propagation.
        '''
        if self.travis:
            return True

        result = self.data_client.transfer(files=[{'scope': 'tests',
                                                   'file': 'i_do_not_exist',
                                                   'rse': 'CERN-PROD_SCRATCHDISK'},
                                                  {'scope': 'tests',
                                                   'file': 'neither_do_i',
                                                   'rse': 'CERN-PROD_SCRATCHDISK'}])

        for _file in result:
            self.assertEqual(_file['errno'], 1)

    def test_stageout_file(self):
        '''
        Single file upload with various combinations of parameters.
        '''
        if self.travis:
            return True

        tmp_fd, tmp_file1 = tempfile.mkstemp()
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()
        tmp_fd, tmp_file2 = tempfile.mkstemp()
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()
        tmp_fd, tmp_file3 = tempfile.mkstemp()
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()
        tmp_fd, tmp_file4 = tempfile.mkstemp()
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()

        result = self.data_client.transfer(files=[{'scope': 'tests',
                                                   'file': tmp_file1,
                                                   'rse': 'CERN-PROD_SCRATCHDISK'},
                                                  {'scope': 'tests',
                                                   'file': tmp_file2,
                                                   'lifetime': 600,
                                                   'rse': 'CERN-PROD_SCRATCHDISK'},
                                                  {'scope': 'tests',
                                                   'file': tmp_file3,
                                                   'lifetime': 600,
                                                   'summary': True,
                                                   'rse': 'CERN-PROD_SCRATCHDISK'},
                                                  {'scope': 'tests',
                                                   'file': tmp_file4,
                                                   'guid': str(uuid.uuid4()),
                                                   'rse': 'CERN-PROD_SCRATCHDISK'}])

        for _file in result:
            self.assertEqual(_file['errno'], 0)

    def test_stageout_file_and_attach(self):
        '''
        Single file upload and attach to dataset.
        '''
        if self.travis:
            return True

        tmp_fd, tmp_file1 = tempfile.mkstemp()
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()
        tmp_fd, tmp_file2 = tempfile.mkstemp()
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()

        result = self.data_client.transfer(files=[{'scope': 'tests',
                                                   'file': tmp_file1,
                                                   'lifetime': 600,
                                                   'rse': 'CERN-PROD_SCRATCHDISK',
                                                   'attach': {'scope': 'tests',
                                                              'name': 'pilot3.tests.test_harvester'}},
                                                  {'scope': 'tests',
                                                   'file': tmp_file2,
                                                   'rse': 'CERN-PROD_SCRATCHDISK',
                                                   'attach': {'scope': 'tests',
                                                              'name': 'pilot3.tests.test_harvester'}}])

        for _file in result:
            self.assertEqual(_file['errno'], 0)

    def test_stageout_file_noregister(self):
        '''
        Single file upload without registering.
        '''
        if self.travis:
            return True

        tmp_fd, tmp_file1 = tempfile.mkstemp()
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()
        tmp_fd, tmp_file2 = tempfile.mkstemp()
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()

        result = self.data_client.transfer(files=[{'scope': 'tests',
                                                   'file': tmp_file1,
                                                   'rse': 'CERN-PROD_SCRATCHDISK',
                                                   'no_register': True},
                                                  {'scope': 'tests',
                                                   'file': tmp_file2,
                                                   'rse': 'CERN-PROD_SCRATCHDISK',
                                                   'no_register': True}])

        for _file in result:
            self.assertEqual(_file['errno'], 0)

    def test_stageout_dir(self):
        '''
        Single file upload.
        '''
        if self.travis:
            return True

        tmp_dir = tempfile.mkdtemp()
        tmp_fd, tmp_file1 = tempfile.mkstemp(dir=tmp_dir)
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()
        tmp_fd, tmp_file2 = tempfile.mkstemp(dir=tmp_dir)
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()
        tmp_fd, tmp_file3 = tempfile.mkstemp(dir=tmp_dir)
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()

        result = self.data_client.transfer(files=[{'scope': 'tests',
                                                   'file': tmp_dir,
                                                   'rse': 'CERN-PROD_SCRATCHDISK'}])

        for _file in result:
            self.assertEqual(_file['errno'], 0)

    def test_stageout_dir_and_attach(self):
        '''
        Single file upload and attach to dataset.
        '''
        if self.travis:
            return True

        tmp_dir = tempfile.mkdtemp()
        tmp_fd, tmp_file1 = tempfile.mkstemp(dir=tmp_dir)
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()
        tmp_fd, tmp_file2 = tempfile.mkstemp(dir=tmp_dir)
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()
        tmp_fd, tmp_file3 = tempfile.mkstemp(dir=tmp_dir)
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()

        result = self.data_client.transfer(files=[{'scope': 'tests',
                                                   'file': tmp_dir,
                                                   'lifetime': 600,
                                                   'rse': 'CERN-PROD_SCRATCHDISK',
                                                   'attach': {'scope': 'tests',
                                                              'name': 'pilot3.tests.test_harvester'}}])

        for _file in result:
            self.assertEqual(_file['errno'], 0)

    def test_stageout_dir_noregister(self):
        '''
        Single file upload without registering.
        '''
        if self.travis:
            return True

        tmp_dir = tempfile.mkdtemp()
        tmp_fd, tmp_file1 = tempfile.mkstemp(dir=tmp_dir)
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()
        tmp_fd, tmp_file2 = tempfile.mkstemp(dir=tmp_dir)
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()
        tmp_fd, tmp_file3 = tempfile.mkstemp(dir=tmp_dir)
        tmp_fdo = os.fdopen(tmp_fd, 'wb')
        tmp_fdo.write(str(random.randint(1, 2**2048)))
        tmp_fdo.close()

        result = self.data_client.transfer(files=[{'scope': 'tests',
                                                   'file': tmp_dir,
                                                   'no_register': True,
                                                   'rse': 'CERN-PROD_SCRATCHDISK'}])

        for _file in result:
            self.assertEqual(_file['errno'], 0)



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Daniel Drizhuk, d.drizhuk@gmail.com, 2017
# - Mario Lassnig, mario.lassnig@cern.ch, 2017
# - Paul Nilsson, paul.nilsson@cern.ch, 2017-2022

import json
import os
import platform
import random
import socket
import ssl
import sys
import urllib.request
import urllib.error
import urllib.parse
import pipes
from collections import namedtuple
from time import sleep, time
from re import findall

from .filehandling import write_file, read_file
from .config import config
from .constants import get_pilot_version
from .container import execute
from pilot.common.errorcodes import ErrorCodes

import logging
logger = logging.getLogger(__name__)
errors = ErrorCodes()

_ctx = namedtuple('_ctx', 'ssl_context user_agent capath cacert')
_ctx.ssl_context = None
_ctx.user_agent = None
_ctx.capath = None
_ctx.cacert = None

# anisyonk: public copy of `_ctx` to avoid logic break since ssl_context is reset inside the request() -- FIXME
# anisyonk: public instance, should be properly initialized by `https_setup()`
# anisyonk: use lightweight class definition instead of namedtuple since tuple is immutable and we don't need/use any tuple features here
ctx = type('ctx', (object,), dict(ssl_context=None, user_agent='Pilot2 client', capath=None, cacert=None))


def _tester(func, *args):
    """
    Tests function ``func`` on arguments and returns first positive.

    >>> _tester(lambda x: x%3 == 0, 1, 2, 3, 4, 5, 6)
    3
    >>> _tester(lambda x: x%3 == 0, 1, 2)
    None

    :param func: function(arg)->boolean
    :param args: other arguments
    :return: something or none
    """
    for arg in args:
        if arg is not None and func(arg):
            return arg

    return None


def capath(args=None):
    """
    Tries to get :abbr:`CA (Certification Authority)` path with certificates.
    Testifies it to be a directory.
    Tries next locations:

    1. :option:`--capath` from arguments
    2. :envvar:`X509_CERT_DIR` from env
    3. Path ``/etc/grid-security/certificates``

    :param args: arguments, parsed by `argparse`
    :returns: `str` -- directory path, or `None`
    """

    return _tester(os.path.isdir,
                   args and args.capath,
                   os.environ.get('X509_CERT_DIR'),
                   '/etc/grid-security/certificates')


def cacert_default_location():
    """
    Tries to get current user ID through `os.getuid`, and get the posix path for x509 certificate.
    :returns: `str` -- posix default x509 path, or `None`
    """
    try:
        return '/tmp/x509up_u%s' % str(os.getuid())
    except AttributeError:
        logger.warning('No UID available? System not POSIX-compatible... trying to continue')
        pass

    return None


def cacert(args=None):
    """
    Tries to get :abbr:`CA (Certification Authority)` certificate or X509 one.
    Testifies it to be a regular file.
    Tries next locations:

    1. :option:`--cacert` from arguments
    2. :envvar:`X509_USER_PROXY` from env
    3. Path ``/tmp/x509up_uXXX``, where ``XXX`` refers to ``UID``

    :param args: arguments, parsed by `argparse`
    :returns: `str` -- certificate file path, or `None`
    """

    return _tester(os.path.isfile,
                   args and args.cacert,
                   os.environ.get('X509_USER_PROXY'),
                   cacert_default_location())


def https_setup(args=None, version=None):
    """
    Sets up the context for future HTTPS requests:

    1. Selects the certificate paths
    2. Sets up :mailheader:`User-Agent`
    3. Tries to create `ssl.SSLContext` for future use (falls back to :command:`curl` if fails)

    :param args: arguments, parsed by `argparse`
    :param str version: pilot version string (for :mailheader:`User-Agent`)
    """

    version = version or get_pilot_version()

    _ctx.user_agent = 'pilot/%s (Python %s; %s %s)' % (version,
                                                       sys.version.split()[0],
                                                       platform.system(),
                                                       platform.machine())
    _ctx.capath = capath(args)
    _ctx.cacert = cacert(args)

    try:
        _ctx.ssl_context = ssl.create_default_context(capath=_ctx.capath,
                                                      cafile=_ctx.cacert)
    except Exception as exc:
        logger.warning(f'SSL communication is impossible due to SSL error: {exc} -- falling back to curl')
        _ctx.ssl_context = None

    # anisyonk: clone `_ctx` to avoid logic break since ssl_context is reset inside the request() -- FIXME
    ctx.capath = _ctx.capath
    ctx.cacert = _ctx.cacert
    ctx.user_agent = _ctx.user_agent

    try:
        ctx.ssl_context = ssl.create_default_context(capath=ctx.capath, cafile=ctx.cacert)
        ctx.ssl_context.load_cert_chain(ctx.cacert)
    except Exception as exc:
        logger.warning(f'Failed to initialize SSL context .. skipped, error: {exc}')


def request(url, data=None, plain=False, secure=True, ipv='IPv6'):
    """
    This function sends a request using HTTPS.
    Sends :mailheader:`User-Agent` and certificates previously being set up by `https_setup`.
    If `ssl.SSLContext` is available, uses `urllib2` as a request processor. Otherwise uses :command:`curl`.

    If ``data`` is provided, encodes it as a URL form data and sends it to the server.

    Treats the request as JSON unless a parameter ``plain`` is `True`.
    If JSON is expected, sends ``Accept: application/json`` header.

    :param string url: the URL of the resource.
    :param dict data: data to send.
    :param boolean plain: if true, treats the response as a plain text.
    :param secure: Boolean (default: True, ie use certificates).
    :param ipv: internet protocol version (string).
    Usage:

    .. code-block:: python
        :emphasize-lines: 2

        https_setup(args, PILOT_VERSION)  # sets up ssl and other stuff
        response = request('https://some.url', {'some':'data'})

    Returns:
        - :keyword:`dict` -- if everything went OK
        - `str` -- if ``plain`` parameter is `True`
        - `None` -- if something went wrong
    """

    _ctx.ssl_context = None  # certificates are not available on the grid, use curl

    # note that X509_USER_PROXY might change during running (in the case of proxy downloads), so
    # we might have to update _ctx
    update_ctx()

    logger.debug(f'server update dictionary = \n{data}')

    # get the filename and strdata for the curl config file
    filename, strdata = get_vars(url, data)
    # write the strdata to file
    writestatus = write_file(filename, strdata)
    # get the config option for the curl command
    dat = get_curl_config_option(writestatus, url, data, filename)

    # loop over internet protocol versions since proper value might be known yet (ie before downloading queuedata)
    ipvs = ['IPv6', 'IPv4'] if ipv == 'IPv6' else ['IPv4']
    if _ctx.ssl_context is None and secure:
        failed = False
        for _ipv in ipvs:
            req, obscure = get_curl_command(plain, dat, _ipv)
            if not req:
                logger.warning('failed to construct valid curl command')
                failed = True
                break
            try:
                status, output, stderr = execute(req, obscure=obscure)
            except Exception as exc:
                logger.warning(f'exception: {exc}')
                failed = True
                break
            else:
                if status == 0:
                    break
                else:
                    logger.warning(f'request failed for IPv={_ipv} ({status}): stdout={output}, stderr={stderr}')
                    continue
        if failed:
            return None

        # return output if plain otherwise return json.loads(output)
        if plain:
            return output
        else:
            try:
                ret = json.loads(output)
            except Exception as exc:
                logger.warning(f'json.loads() failed to parse output={output}: {exc}')
                return None
            else:
                return ret
    else:
        req = execute_urllib(url, data, plain, secure)
        context = _ctx.ssl_context if secure else None

        ec, output = get_urlopen_output(req, context)
        if ec:
            return None

        return output.read() if plain else json.load(output)


def update_ctx():
    """
    Update the ctx object in case X509_USER_PROXY has been updated.
    """

    x509 = os.environ.get('X509_USER_PROXY', _ctx.cacert)
    if x509 != _ctx.cacert and os.path.exists(x509):
        _ctx.cacert = x509


def get_curl_command(plain, dat, ipv):
    """
    Get the curl command.

    :param plain:
    :param dat: curl config option (string).
    :param ipv: internet protocol version (string).
    :return: curl command (string), sensitive string to be obscured before dumping to log (string).
    """

    auth_token_content = ''
    auth_token = os.environ.get('OIDC_AUTH_TOKEN', os.environ.get('PANDA_AUTH_TOKEN', None))  # file name of the token
    auth_origin = os.environ.get('OIDC_AUTH_ORIGIN', os.environ.get('PANDA_AUTH_ORIGIN', None))  # origin of the token (panda_dev.pilot)

    command = 'curl'
    if ipv == 'IPv4':
        command += ' -4'
    if auth_token and auth_origin:
        # curl --silent --capath
        # /cvmfs/atlas.cern.ch/repo/ATLASLocalRootBase/etc/grid-security-emi/certificates --compressed
        # -H "Authorization: Bearer <contents of PANDA_AUTH_TOKEN>" -H "Origin: <PANDA_AUTH_VO>"
        path = locate_token(auth_token)
        auth_token_content = ""
        if os.path.exists(path):
            auth_token_content = read_file(path)
            if not auth_token_content:
                logger.warning(f'failed to read file {path}')
                return None, ''
        else:
            logger.warning(f'path does not exist: {path}')
            return None, ''
        if not auth_token_content:
            logger.warning('OIDC_AUTH_TOKEN/PANDA_AUTH_TOKEN content could not be read')
            return None, ''
        req = f'{command} -sS --compressed --connect-timeout {config.Pilot.http_connect_timeout} ' \
              f'--max-time {config.Pilot.http_maxtime} '\
              f'--capath {pipes.quote(_ctx.capath or "")} ' \
              f'-H "Authorization: Bearer {pipes.quote(auth_token_content)}" ' \
              f'-H {pipes.quote("Accept: application/json") if not plain else ""} ' \
              f'-H "Origin: {pipes.quote(auth_origin)}" {dat}'
    else:
        req = f'{command} -sS --compressed --connect-timeout {config.Pilot.http_connect_timeout} ' \
              f'--max-time {config.Pilot.http_maxtime} '\
              f'--capath {pipes.quote(_ctx.capath or "")} ' \
              f'--cert {pipes.quote(_ctx.cacert or "")} ' \
              f'--cacert {pipes.quote(_ctx.cacert or "")} ' \
              f'--key {pipes.quote(_ctx.cacert or "")} '\
              f'-H {pipes.quote("User-Agent: %s" % _ctx.user_agent)} ' \
              f'-H {pipes.quote("Accept: application/json") if not plain else ""} {dat}'

    #logger.info('request: %s', req)
    return req, auth_token_content


def locate_token(auth_token):
    """
    Locate the token file.

    :param auth_token: file name of token (string).
    :return: path to token (string).
    """

    _primary = os.path.dirname(os.environ.get('OIDC_AUTH_DIR', os.environ.get('PANDA_AUTH_DIR', os.environ.get('X509_USER_PROXY', ''))))
    paths = [os.path.join(_primary, auth_token),
             os.path.join(os.environ.get('PILOT_SOURCE_DIR', ''), auth_token),
             os.path.join(os.environ.get('PILOT_WORK_DIR', ''), auth_token)]
    path = ""
    for _path in paths:
        logger.debug(f'looking for {_path}')
        if os.path.exists(_path):
            path = _path
            break

    if path == "":
        logger.info(f'did not find any local token file ({auth_token}) in paths={paths}')

    return path


def get_vars(url, data):
    """
    Get the filename and strdata for the curl config file.

    :param url: URL (string).
    :param data: data to be written to file (dictionary).
    :return: filename (string), strdata (string).
    """

    strdata = ""
    for key in data:
        strdata += 'data="%s"\n' % urllib.parse.urlencode({key: data[key]})
    jobid = ''
    if 'jobId' in list(data.keys()):
        jobid = '_%s' % data['jobId']

    # write data to temporary config file
    filename = '%s/curl_%s%s.config' % (os.getenv('PILOT_HOME'), os.path.basename(url), jobid)

    return filename, strdata


def get_curl_config_option(writestatus, url, data, filename):
    """
    Get the curl config option.

    :param writestatus: status of write_file call (Boolean).
    :param url: URL (string).
    :param data: data structure (dictionary).
    :param filename: file name of config file (string).
    :return: config option (string).
    """

    if not writestatus:
        logger.warning('failed to create curl config file (will attempt to urlencode data directly)')
        dat = pipes.quote(url + '?' + urllib.parse.urlencode(data) if data else '')
    else:
        dat = f'--config {filename} {url}'

    return dat


def execute_urllib(url, data, plain, secure):
    """
    Execute the request using urllib.

    :param url: URL (string).
    :param data: data structure
    :return: urllib request structure.
    """

    req = urllib.request.Request(url, urllib.parse.urlencode(data))
    if not plain:
        req.add_header('Accept', 'application/json')
    if secure:
        req.add_header('User-Agent', _ctx.user_agent)

    return req


def get_urlopen_output(req, context):
    """
    Get the output from the urlopen request.

    :param req:
    :param context:
    :return: ec (int), output (string).
    """

    exitcode = -1
    output = ""
    try:
        output = urllib.request.urlopen(req, context=context)
    except urllib.error.HTTPError as exc:
        logger.warning(f'server error ({exc.code}): {exc.read()}')
    except urllib.error.URLError as exc:
        logger.warning(f'connection error: {exc.reason}')
    else:
        exitcode = 0

    return exitcode, output


def send_update(update_function, data, url, port, job=None, ipv='IPv6'):
    """
    Send the update to the server using the given function and data.

    :param update_function: 'updateJob' or 'updateWorkerPilotStatus' (string).
    :param data: data (dictionary).
    :param url: server url (string).
    :param port: server port (string).
    :param job: job object.
    :param ipv: internet protocol version, IPv4 or IPv6 (string).
    :return: server response (dictionary).
    """

    time_before = int(time())
    max_attempts = 10
    attempt = 0
    done = False
    res = None

    if os.environ.get('REACHED_MAXTIME', None) and update_function == 'updateJob':
        data['state'] = 'failed'
        if job:
            job.state = 'failed'
            msg = 'the max batch system time limit has been reached'
            logger.warning(msg)
            job.piloterrorcodes, job.piloterrordiags = errors.add_error_code(errors.REACHEDMAXTIME, msg=msg)
            add_error_codes(data, job)

    while attempt < max_attempts and not done:
        logger.info(f'server update attempt {attempt + 1}/{max_attempts}')

        # get the URL for the PanDA server from pilot options or from config
        try:
            pandaserver = get_panda_server(url, port)
        except Exception as exc:
            logger.warning(f'exception caught in get_panda_server(): {exc}')
            sleep(5)
            attempt += 1
            continue
        # send the heartbeat
        try:
            res = request(f'{pandaserver}/server/panda/{update_function}', data=data, ipv=ipv)
        except Exception as exc:
            logger.warning(f'exception caught in https.request(): {exc}')
        else:
            if res is not None:
                done = True
            txt = f'server {update_function} request completed in {int(time()) - time_before}s'
            if job:
                txt += f' for job {job.jobid}'
            logger.info(txt)
            # hide sensitive info
            pilotsecrets = ''
            if res and 'pilotSecrets' in res:
                pilotsecrets = res['pilotSecrets']
                res['pilotSecrets'] = '********'
            logger.info(f'server responded with: res = {res}')
            if pilotsecrets:
                res['pilotSecrets'] = pilotsecrets

        attempt += 1
    return res


def get_panda_server(url, port, update_server=True):
    """
    Get the URL for the PanDA server.
    The URL will be randomized if the server can be contacted (otherwise fixed).

    :param url: URL string, if set in pilot option (port not included).
    :param port: port number, if set in pilot option (int).
    :param update_server: True if the server can be contacted (Boolean).
    :return: full URL (either from pilot options or from config file).
    """

    if url != '':
        parsedurl = url.split('://')
        scheme = None
        if len(parsedurl) == 2:
            scheme = parsedurl[0]
            loc = parsedurl[1]
        else:
            loc = parsedurl[0]

        parsedloc = loc.split(':')
        loc = parsedloc[0]

        # if a port is provided in the url, then override the port argument
        if len(parsedloc) == 2:
            port = parsedloc[1]
        # default scheme to https
        if not scheme:
            scheme = "https"
        portstr = f":{port}" if port else ""
        pandaserver = f"{scheme}://{loc}{portstr}"
    else:
        pandaserver = config.Pilot.pandaserver
        if not pandaserver.startswith('http'):
            pandaserver = 'https://' + pandaserver

    if not update_server:
        return pandaserver

    # add randomization for PanDA server
    default = 'pandaserver.cern.ch'
    if default in pandaserver:
        rnd = random.choice([socket.getfqdn(vv) for vv in set([v[-1][0] for v in socket.getaddrinfo(default, 25443, socket.AF_INET)])])
        pandaserver = pandaserver.replace(default, rnd)
        logger.debug(f'updated {default} to {pandaserver}')

    return pandaserver


def add_error_codes(data, job):
    """
    Add error codes to data structure.

    :param data: data dictionary.
    :param job: job object.
    :return:
    """

    # error codes
    pilot_error_code = job.piloterrorcode
    pilot_error_codes = job.piloterrorcodes
    if pilot_error_codes != []:
        logger.warning(f'pilotErrorCodes = {pilot_error_codes} (will report primary/first error code)')
        data['pilotErrorCode'] = pilot_error_codes[0]
    else:
        data['pilotErrorCode'] = pilot_error_code

    # add error info
    pilot_error_diag = job.piloterrordiag
    pilot_error_diags = job.piloterrordiags
    if pilot_error_diags != []:
        logger.warning(f'pilotErrorDiags = {pilot_error_diags} (will report primary/first error diag)')
        data['pilotErrorDiag'] = pilot_error_diags[0]
    else:
        data['pilotErrorDiag'] = pilot_error_diag
    data['transExitCode'] = job.transexitcode
    data['exeErrorCode'] = job.exeerrorcode
    data['exeErrorDiag'] = job.exeerrordiag


def get_server_command(url, port, cmd='getJob'):
    """
    Prepare the getJob server command.

    :param url: PanDA server URL (string)
    :param port: PanDA server port
    :return: full server command (URL string)
    """

    if url != "":
        port_pattern = '.:([0-9]+)'
        if not findall(port_pattern, url):
            url = url + ':%s' % port
        else:
            logger.debug(f'URL already contains port: {url}')
    else:
        url = config.Pilot.pandaserver
    if url == "":
        logger.fatal('PanDA server url not set (either as pilot option or in config file)')
    elif not url.startswith("http"):
        url = 'https://' + url
        logger.warning('detected missing protocol in server url (added)')

    # randomize server name
    url = get_panda_server(url, port)
    return f'{url}/server/panda/{cmd}'



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2017-2023

import logging
import os
import traceback

from pilot.common.exception import FileHandlingFailure
from pilot.util import https
from pilot.util.config import config
from pilot.util.container import execute
from pilot.util.filehandling import write_file

logger = logging.getLogger(__name__)


def get_distinguished_name():
    """
    Get the user DN.
    Note: the DN is also sent by the server to the pilot in the job description (produserid).

    :return: User DN (string).
    """

    dn = ""
    executable = 'arcproxy -i subject'
    exit_code, stdout, stderr = execute(executable)
    if exit_code != 0 or "ERROR:" in stderr:
        logger.warning("arcproxy failed: ec=%d, stdout=%s, stderr=%s" % (exit_code, stdout, stderr))

        if "command not found" in stderr or "Can not find certificate file" in stderr:
            logger.warning("arcproxy experienced a problem (will try voms-proxy-info instead)")

            # Default to voms-proxy-info
            exit_code, stdout, stderr = vomsproxyinfo(options='-subject', mute=True)

    if exit_code == 0:
        dn = stdout
        logger.info('DN = %s' % dn)
        cn = "/CN=proxy"
        if not dn.endswith(cn):
            logger.info("DN does not end with %s (will be added)" % cn)
            dn += cn

    else:
        logger.warning("user=self set but cannot get proxy: %d, %s" % (exit_code, stdout))

    return dn


def vomsproxyinfo(options='-all', mute=False, path=''):
    """
    Execute voms-proxy-info with the given options.

    :param options: command options (string).
    :param mute: should command output be printed (mute=False).
    :param path: use given path if specified for proxy (string).
    :return: exit code (int), stdout (string), stderr (string).
    """

    executable = f'voms-proxy-info {options}'
    if path:
        executable += f' --file={path}'
    exit_code, stdout, stderr = execute(executable)
    if not mute:
        logger.info(stdout + stderr)

    return exit_code, stdout, stderr


def get_proxy(proxy_outfile_name, voms_role):
    """
    Download and store a proxy.
    E.g. on read-only file systems (K8), the default path is not available, in which case the new proxy
    will be stored in the workdir (return the updated path).

    :param proxy_outfile_name: specify the file to store proxy (string).
    :param voms_role: what proxy (role) to request, e.g. 'atlas' (string).
    :return: result (Boolean), updated proxy path (string).
    """
    try:
        # it assumes that https_setup() was done already
        url = os.environ.get('PANDA_SERVER_URL', config.Pilot.pandaserver)

        pilot_user = os.environ.get('PILOT_USER', 'generic').lower()
        user = __import__('pilot.user.%s.proxy' % pilot_user, globals(), locals(), [pilot_user], 0)
        data = user.getproxy_dictionary(voms_role)

        res = https.request('{pandaserver}/server/panda/getProxy'.format(pandaserver=url), data=data)
        if res is None:
            logger.error(f"unable to get proxy with role '{voms_role}' from panda server")
            return False, proxy_outfile_name

        if res['StatusCode'] != 0:
            logger.error(f"panda server returned: \'{res['errorDialog']}\' for proxy role \'{voms_role}\'")
            return False, proxy_outfile_name

        proxy_contents = res['userProxy']

    except Exception as exc:
        logger.error(f"Get proxy from panda server failed: {exc}, {traceback.format_exc()}")
        return False, proxy_outfile_name
    else:
        # dump voms-proxy-info -all to log
        if res and res['StatusCode'] == 0:
            _, _, _ = vomsproxyinfo(options='-all', path=proxy_outfile_name)

    def create_file(filename, contents):
        """
        Internally used helper function to create proxy file.
        """
        _file = os.open(filename, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o600)
        os.close(_file)
        return write_file(filename, contents, mute=False)  # returns True on success

    result = False
    try:
        # pre-create empty proxy file with secure permissions. Prepare it for write_file() which can not
        # set file permission mode, it will writes to the existing file with correct permissions.
        result = create_file(proxy_outfile_name, proxy_contents)
    except (IOError, OSError, FileHandlingFailure) as exc:
        logger.error(f"exception caught:\n{exc},\ntraceback: {traceback.format_exc()}")
        if 'Read-only file system' in exc:
            proxy_outfile_name = os.path.join(os.getenv('PILOT_HOME'), os.path.basename(proxy_outfile_name))  # e.g. '/path/x509up_u25606_prod-unified.proxy'
            logger.info(f'attempting writing proxy to alternative path: {proxy_outfile_name}')
            try:  # can we bypass a problem with read-only file systems by writing the proxy to the pilot home dir instead?
                result = create_file(proxy_outfile_name, proxy_contents)
            except (IOError, OSError, FileHandlingFailure) as exc:
                logger.error(f"exception caught:\n{exc},\ntraceback: {traceback.format_exc()}")
            else:
                logger.debug('updating X509_USER_PROXY to alternative path {path} (valid until end of current job)')
                os.environ['X509_USER_PROXY'] = proxy_outfile_name

    return result, proxy_outfile_name


def create_cert_files(from_proxy, workdir):
    """
    Create cert/key pem files from given proxy and store in workdir.
    These files are needed for communicating with logstash server.

    :param from_proxy: path to proxy file (string).
    :param workdir: work directory (string).
    :return: path to crt.pem (string), path to key.pem (string).
    """

    _files = [os.path.join(workdir, 'crt.pem'), os.path.join(workdir, 'key.pem')]
    if os.path.exists(_files[0]) and os.path.exists(_files[1]):
        return _files[0], _files[1]

    cmds = [f'openssl pkcs12 -in {from_proxy} -out {_files[0]} -clcerts -nokeys',
            f'openssl pkcs12 -in {from_proxy} -out {_files[1]} -nocerts -nodes']

    counter = 0
    for cmd in cmds:
        ec, stdout, stderr = execute(cmd)
        if ec:
            logger.warning(f'cert command failed: {stdout}, {stderr}')
            return '', ''
        else:
            logger.debug(f'produced key/cert file: {_files[counter]}')
            counter += 1

    return _files[0], _files[1]



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Mario Lassnig, mario.lassnig@cern.ch, 2016
# - Paul Nilsson, paul.nilsson@cern.ch, 2018-2019

import functools
import signal
from collections import namedtuple
from os import environ

from pilot.util.constants import SUCCESS, FAILURE

import logging
logger = logging.getLogger(__name__)


def interrupt(args, signum, frame):
    try:
        logger.info('caught signal: %s' % [v for v, k in signal.__dict__.iteritems() if k == signum][0])
    except Exception:
        logger.info('caught signal: %s' % [v for v, k in list(signal.__dict__.items()) if k == signum][0])
    args.graceful_stop.set()


def run(args):
    """
    Main execution function for the event service workflow on HPCs (Yoda-Droid).

    :param args: pilot arguments.
    :returns: traces object.
    """

    try:
        logger.info('setting up signal handling')
        signal.signal(signal.SIGINT, functools.partial(interrupt, args))

        logger.info('setting up tracing')
        traces = namedtuple('traces', ['pilot'])
        traces.pilot = {'state': SUCCESS,
                        'nr_jobs': 0}

        if args.hpc_resource == '':
            logger.critical('hpc resource not specified, cannot continue')
            traces.pilot['state'] = FAILURE
            return traces

        # get the resource reference
        resource = __import__('pilot.resource.%s' % args.hpc_resource, globals(), locals(), [args.hpc_resource], 0)  # Python 2/3

        # example usage:
        logger.info('setup for resource %s: %s' % (args.hpc_resource, str(resource.get_setup())))

        # are we Yoda or Droid?
        if environ.get('SOME_ENV_VARIABLE', '') == 'YODA':
            yodadroid = __import__('pilot.eventservice.yoda')
        else:
            yodadroid = __import__('pilot.eventservice.droid')
        yodadroid.run()

    except Exception as e:
        logger.fatal('exception caught: %s' % e)

    return traces



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Wen Guan, wen.guan@cern.ch, 2018

import json
import logging
import os
import socket
import sys
import time

from pilot.eventservice.communicationmanager.communicationmanager import CommunicationRequest, CommunicationResponse, CommunicationManager
from pilot.util.https import https_setup
from pilot.util.timing import time_stamp

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest

logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
logger = logging.getLogger(__name__)


https_setup(None, None)


def check_env():
    """
    Function to check whether cvmfs is available.
    To be used to decide whether to skip some test functions.

    :returns True: if cvmfs is available. Otherwise False.
    """
    return os.path.exists('/cvmfs/atlas.cern.ch/repo/')


class TestESCommunicationRequestResponse(unittest.TestCase):
    """
    Unit tests for event service communicator Request and Response.
    """

    def test_communicator_request(self):
        """
        Make sure that es message thread works as expected.
        """
        req_attrs = {'request_type': CommunicationRequest.RequestType.RequestJobs,
                     'num_jobs': 1, 'post_hook': None, 'response': None}
        req_job = CommunicationRequest(req_attrs)
        self.assertEqual(req_job.request_type, CommunicationRequest.RequestType.RequestJobs)

        req_attrs = {'request_type': CommunicationRequest.RequestType.RequestEvents,
                     'num_event_ranges': 1, 'post_hook': None, 'response': None}
        req_events = CommunicationRequest(req_attrs)
        self.assertEqual(req_events.request_type, CommunicationRequest.RequestType.RequestEvents)

        req_attrs = {'request_type': CommunicationRequest.RequestType.UpdateEvents,
                     'output_files': None, 'post_hook': None, 'response': None}
        req_output = CommunicationRequest(req_attrs)
        self.assertEqual(req_output.request_type, CommunicationRequest.RequestType.UpdateEvents)

        resp_attrs = {'status': 0, 'content': None, 'exception': None}
        resp = CommunicationResponse(resp_attrs)
        self.assertEqual(resp.status, 0)


class TestESCommunicationManagerPanda(unittest.TestCase):
    """
    Unit tests for event service communicator manager.
    """

    @unittest.skipIf(not check_env(), "No CVMFS")
    def test_communicator_manager(self):
        """
        Make sure that es communicator manager thread works as expected.
        """
        communicator_manager = None
        try:
            args = {'workflow': 'eventservice_hpc',
                    'queue': 'BNL_CLOUD_MCORE',
                    'site': 'BNL_CLOUD_MCORE',
                    'port': 25443,
                    'url': 'https://aipanda007.cern.ch',
                    'job_label': 'ptest',
                    'pilot_user': 'ATLAS',
                    'node': socket.getfqdn(),
                    'mem': 16000,
                    'disk_space': 160000,
                    'working_group': '',
                    'cpu': 2601.0,
                    'info': None}

            communicator_manager = CommunicationManager()
            communicator_manager.start()
            self.assertTrue(communicator_manager.is_alive())

            jobs = communicator_manager.get_jobs(njobs=2, args=args)
            self.assertEqual(len(jobs), 2)

            jobs = communicator_manager.get_jobs(njobs=1, args=args)
            self.assertEqual(len(jobs), 1)

            job_list = []
            for job in jobs:
                job_data = {'node': socket.getfqdn(),
                            'pilotErrorCode': 0,
                            'startTime': time.time(),
                            'jobMetrics': 'coreCount=8',
                            'schedulerID': 'unknown',
                            'timestamp': time_stamp(),
                            'exeErrorCode': 0,
                            'pilotID': 'unknown|PR|2.0.0 (80)',
                            'transExitCode': 0,
                            'pilotErrorDiag': '',
                            'exeErrorDiag': ''}
                job_data['jobId'] = job['PandaID']
                job_data['siteName'] = 'BNL_CLOUD_MCORE'
                job_data['state'] = 'running'
                job_data['attemptNr'] = job['attemptNr'] + 1
                job_list.append(job_data)
            status = communicator_manager.update_jobs(jobs=job_list)
            self.assertEqual(status[0], True)

            events = communicator_manager.get_event_ranges(num_event_ranges=1, job=jobs[0])
            self.assertEqual(len(events), 1)

            for event in events:
                event_range_status = {"errorCode": 1220, "eventRangeID": event['eventRangeID'], "eventStatus": 'failed'}
                event_range_message = {'version': 0, 'eventRanges': json.dumps(event_range_status)}
                res = communicator_manager.update_events(update_events=event_range_message)
                self.assertEqual(res['StatusCode'], 0)

            events = communicator_manager.get_event_ranges(num_event_ranges=2, job=jobs[0])
            self.assertEqual(len(events), 2)

            update_events = []
            for event in events:
                event_range = {"eventRangeID": event['eventRangeID'], "eventStatus": 'finished'}
                update_events.append(event_range)
            event_range_status = [{"zipFile": {"numEvents": len(update_events),
                                               "objstoreID": 1318,
                                               "adler32": '000000',
                                               "lfn": 'test_file',
                                               "fsize": 100,
                                               "pathConvention": 1000},
                                   "eventRanges": update_events}]

            event_range_message = {'version': 1, 'eventRanges': json.dumps(event_range_status)}
            res = communicator_manager.update_events(update_events=event_range_message)
            self.assertEqual(res['StatusCode'], 0)

            communicator_manager.stop()
            time.sleep(2)
            self.assertFalse(communicator_manager.is_alive())
        except Exception as ex:
            if communicator_manager:
                communicator_manager.stop()
            raise ex



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2018-2022

import os
import os.path
import socket

from pilot.common.exception import FileHandlingFailure
from pilot.util.config import config
from pilot.util.filehandling import write_json, touch, remove, read_json, get_checksum_value
from pilot.util.timing import time_stamp

import logging
logger = logging.getLogger(__name__)


def dump(obj):
    """
    function for debugging - dumps object to sysout
    """
    for attr in dir(obj):
        print(f"obj.{attr} = {getattr(obj, attr)}")


def is_harvester_mode(args):
    """
    Determine if the pilot is running in Harvester mode.
    :param args: Pilot arguments object.
    :return: Boolean.
    """

    if (args.harvester_workdir != '' or args.harvester_datadir != '') and not args.update_server:
        harvester = True
    elif (args.harvester_eventstatusdump != '' or args.harvester_workerattributes != '') and not args.update_server:
        harvester = True
    elif ('HARVESTER_ID' in os.environ or 'HARVESTER_WORKER_ID' in os.environ) and args.harvester_submitmode.lower() == 'push':
        harvester = True
    else:
        harvester = False

    return harvester


def get_job_request_file_name():
    """
    Return the name of the job request file as defined in the pilot config file.

    :return: job request file name.
    """

    return os.path.join(os.environ['PILOT_HOME'], config.Harvester.job_request_file)


def remove_job_request_file():
    """
    Remove an old job request file when it is no longer needed.

    :return:
    """

    path = get_job_request_file_name()
    if os.path.exists(path):
        if remove(path) == 0:
            logger.info(f'removed {path}')
    else:
        logger.debug('there is no job request file')


def request_new_jobs(njobs=1):
    """
    Inform Harvester that the pilot is ready to process new jobs by creating a job request file with the desired
    number of jobs.

    :param njobs: Number of jobs. Default is 1 since on grids and clouds the pilot does not know how many jobs it can
    process before it runs out of time.
    :return:
    """

    path = get_job_request_file_name()
    dictionary = {'nJobs': njobs}

    # write it to file
    try:
        write_json(path, dictionary)
    except FileHandlingFailure:
        raise FileHandlingFailure


def kill_worker():
    """
    Create (touch) a kill_worker file in the pilot launch directory.
    This file will let Harverster know that the pilot has finished.

    :return:
    """

    touch(os.path.join(os.environ['PILOT_HOME'], config.Harvester.kill_worker_file))


def get_initial_work_report():
    """
    Prepare the work report dictionary.
    Note: the work_report should also contain all fields defined in parse_jobreport_data().

    :return: work report dictionary.
    """

    hostname = os.environ.get('PAMDA_HOSTNAME', socket.gethostname())
    work_report = {'jobStatus': 'starting',
                   'messageLevel': logging.getLevelName(logger.getEffectiveLevel()),
                   'cpuConversionFactor': 1.0,
                   'cpuConsumptionTime': '',
                   'node': hostname,
                   'workdir': '',
                   'timestamp': time_stamp(),
                   'endTime': '',
                   'transExitCode': 0,
                   'pilotErrorCode': 0,  # only add this in case of failure?
                   }

    return work_report


def get_event_status_file(args):
    """
    Return the name of the event_status.dump file as defined in the pilot config file
    and from the pilot arguments.

    :param args: Pilot arguments object.
    :return: event staus file name.
    """

    logger.debug('config.Harvester.__dict__ : {0}'.format(config.Harvester.__dict__))

    if args.harvester_workdir != '':
        work_dir = args.harvester_workdir
    else:
        work_dir = os.environ['PILOT_HOME']
    event_status_file = config.Harvester.stageoutnfile
    event_status_file = os.path.join(work_dir, event_status_file)
    logger.debug('event_status_file = {}'.format(event_status_file))

    return event_status_file


def get_worker_attributes_file(args):
    """
    Return the name of the worker attributes file as defined in the pilot config file
    and from the pilot arguments.

    :param args: Pilot arguments object.
    :return: worker attributes file name.
    """

    logger.debug('config.Harvester.__dict__ : {0}'.format(config.Harvester.__dict__))

    if args.harvester_workdir != '':
        work_dir = args.harvester_workdir
    else:
        work_dir = os.environ['PILOT_HOME']
    worker_attributes_file = config.Harvester.workerattributesfile
    worker_attributes_file = os.path.join(work_dir, worker_attributes_file)
    logger.debug('worker_attributes_file = {}'.format(worker_attributes_file))

    return worker_attributes_file


def findfile(path, name):
    """
    find the first instance of file in the directory tree

    :param path: directory tree to search
    :param name: name of the file to search

    :return: the path to the first instance of the file
    """

    for root, dirs, files in os.walk(path):
        if name in files:
            return os.path.join(root, name)
    return ''


def publish_stageout_files(job, event_status_file):
    """
    Publishing of work report to file.
    The work report dictionary should contain the fields defined in get_initial_work_report().

    :param args: Pilot arguments object.
    :param job: job object.
    :param event status file name:

    :return: Boolean. status of writing the file information to a json
    """

    # get the harvester workdir from the event_status_file
    work_dir = os.path.dirname(event_status_file)

    out_file_report = {}
    out_file_report[job.jobid] = []

    # first look at the logfile information (logdata) from the FileSpec objects
    for fspec in job.logdata:
        logger.debug("File {} will be checked and declared for stage out".format(fspec.lfn))
        # find the first instance of the file
        filename = os.path.basename(fspec.surl)
        path = findfile(work_dir, filename)
        logger.debug("Found File {} at path - {}".format(fspec.lfn, path))
        #
        file_desc = {}
        file_desc['type'] = fspec.filetype
        file_desc['path'] = path
        file_desc['guid'] = fspec.guid
        file_desc['fsize'] = fspec.filesize
        file_desc['chksum'] = get_checksum_value(fspec.checksum)
        logger.debug("File description - {} ".format(file_desc))
        out_file_report[job.jobid].append(file_desc)

    # Now look at the output file(s) information (outdata) from the FileSpec objects
    for fspec in job.outdata:
        logger.debug("File {} will be checked and declared for stage out".format(fspec.lfn))
        if fspec.status != 'transferred':
            logger.debug('will not add the output file to the json since it was not produced or transferred')
        else:
            # find the first instance of the file
            filename = os.path.basename(fspec.surl)
            path = findfile(work_dir, filename)
            if not path:
                logger.warning('file %s was not found - will not be added to json')
            else:
                logger.debug("Found File {} at path - {}".format(fspec.lfn, path))
                #
                file_desc = {}
                file_desc['type'] = fspec.filetype
                file_desc['path'] = path
                file_desc['guid'] = fspec.guid
                file_desc['fsize'] = fspec.filesize
                file_desc['chksum'] = get_checksum_value(fspec.checksum)
                logger.debug("File description - {} ".format(file_desc))
                out_file_report[job.jobid].append(file_desc)

    if out_file_report[job.jobid]:
        if write_json(event_status_file, out_file_report):
            logger.debug('Stagout declared in: {0}'.format(event_status_file))
            logger.debug('Report for stageout: {}'.format(out_file_report))
            return True
        else:
            logger.debug('Failed to declare stagout in: {0}'.format(event_status_file))
            return False
    else:
        logger.debug('No Report for stageout')
        return False


def publish_work_report(work_report=None, worker_attributes_file="worker_attributes.json"):
    """
    Publishing of work report to file.
    The work report dictionary should contain the fields defined in get_initial_work_report().

    :param work_report: work report dictionary.
    :param worker_attributes_file:
    :raises FileHandlingFailure: in case of IOError.
    :return: True or False
    """

    if work_report:
        try:
            work_report['timestamp'] = time_stamp()
            if "outputfiles" in work_report:
                del (work_report["outputfiles"])
            if "inputfiles" in work_report:
                del (work_report["inputfiles"])
            if "xml" in work_report:
                del (work_report["xml"])
            if write_json(worker_attributes_file, work_report):
                logger.info("work report published: {0}".format(work_report))
                return True
            else:
                logger.error("work report publish failed: {0}".format(work_report))
                return False
        except IOError:
            logger.error("job report copy failed")
            return False
        except Exception as e:
            logger.error("write json file failed: {0}".format(e))
            return False
    else:
        # No work_report return False
        return False


def publish_job_report(job, args, job_report_file="jobReport.json"):
    """
    Copy job report file to make it accessible by Harvester. Shrink job report file.

    :param job: job object.
    :param args: Pilot arguments object.
    :param job_report_file: name of job report (string).
    :raises FileHandlingFailure: in case of IOError.
    :return True or False
    """

    src_file = os.path.join(job.workdir, job_report_file)
    dst_file = os.path.join(args.harvester_workdir, job_report_file)

    try:
        logger.info(
            "copy of payload report [{0}] to access point: {1}".format(job_report_file, args.harvester_workdir))
        # shrink jobReport
        job_report = read_json(src_file)
        if 'executor' in job_report:
            for executor in job_report['executor']:
                if 'logfileReport' in executor:
                    executor['logfileReport'] = {}

        if write_json(dst_file, job_report):
            return True
        else:
            return False

    except IOError:
        logger.error("job report copy failed")
        return False


def parse_job_definition_file(filename):
    """
    This function parses the Harvester job definition file and re-packages the job definition dictionaries.
    The format of the Harvester job definition dictionary is:
    dict = { job_id: { key: value, .. }, .. }
    The function returns a list of these dictionaries each re-packaged as
    dict = { key: value } (where the job_id is now one of the key-value pairs: 'jobid': job_id)

    :param filename: file name (string).
    :return: list of job definition dictionaries.
    """

    job_definitions_list = []

    # re-package dictionaries
    job_definitions_dict = read_json(filename)
    if job_definitions_dict:
        for job_id in job_definitions_dict:
            res = {'jobid': job_id}
            res.update(job_definitions_dict[job_id])
            job_definitions_list.append(res)

    return job_definitions_list



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Wen Guan, wen.guan@cern.ch, 2018

## THIS FILE IS DEPRECATED AND CAN BE REMOVED LATER (anisyonk)

import logging
logger = logging.getLogger(__name__)


sched2ddm_activity_map = {'es_events': 'pw',
                          'es_failover': 'pw',
                          'es_events_read': 'pr'}


def get_ddm_activity(activity):
    """
    Map schedconf activities to corresponding ddm activities

    :param activity: schedconf activity as string
    :return: ddm activity as string
    """
    if activity in sched2ddm_activity_map:
        ddm_activity = sched2ddm_activity_map[activity]
        logger.info("Schedconf activity %s is mapped to ddm activity: %s" % (activity, ddm_activity))
        return ddm_activity
    return activity



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2021

#import re

#import logging

#logger = logging.getLogger(__name__)


def jobparams_prefiltering(value):
    """
    Perform pre-filtering of raw job parameters to avoid problems with especially quotation marks.
    The function can extract some fields from the job parameters to be put back later after actual filtering.

    E.g. ' --athenaopts "HITtoRDO:--nprocs=$ATHENA_CORE_NUMBER" ' will otherwise become
    ' --athenaopts 'HITtoRDO:--nprocs=$ATHENA_CORE_NUMBER' ' which will prevent the environmental variable to be unfolded.

    :param value: job parameters (string).
    :return: list of fields excluded from job parameters (list), updated job parameters (string).
    """

    exclusions = {}

    # Add regex patterns here

    return exclusions, value


def jobparams_postfiltering(value, exclusions={}):
    """
    Perform post-filtering of raw job parameters.
    Any items in the optional exclusion list will be added (space separated) at the end of the job parameters.

    :param value: job parameters (string).
    :param optional exclusions: exlusions dictionary from pre-filtering function (dictionary).
    :return: updated job parameters (string).
    """

    return value



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Mario Lassnig, mario.lassnig@cern.ch, 2016-2017
# - Daniel Drizhuk, d.drizhuk@gmail.com, 2017
# - Paul Nilsson, paul.nilsson@cern.ch, 2017-2022

from __future__ import print_function  # Python 2 (2to3 complains about this)
from __future__ import absolute_import

import argparse
import logging
import os
import sys
import threading
import time
from os import getcwd, chdir, environ
from os.path import exists, join
from shutil import rmtree

from pilot.common.errorcodes import ErrorCodes
from pilot.common.exception import PilotException
from pilot.info import infosys
from pilot.util.auxiliary import pilot_version_banner, shell_exit_code
from pilot.util.constants import SUCCESS, FAILURE, ERRNO_NOJOBS, PILOT_START_TIME, PILOT_END_TIME, get_pilot_version, \
    SERVER_UPDATE_NOT_DONE, PILOT_MULTIJOB_START_TIME
from pilot.util.filehandling import get_pilot_work_dir, mkdirs, establish_logging
from pilot.util.harvester import is_harvester_mode
from pilot.util.https import get_panda_server, https_setup, send_update
from pilot.util.timing import add_to_pilot_timing

errors = ErrorCodes()


def main():
    """
    Main function of PanDA Pilot 2.
    Prepare for and execute the requested workflow.

    :return: exit code (int).
    """

    # get the logger
    logger = logging.getLogger(__name__)

    # print the pilot version
    pilot_version_banner()

    # define threading events
    args.graceful_stop = threading.Event()
    args.abort_job = threading.Event()
    args.job_aborted = threading.Event()

    # define useful variables
    args.retrieve_next_job = True  # go ahead and download a new job
    args.signal = None  # to store any incoming signals
    args.signal_counter = 0  # keep track of number of received kill signal (suicide counter)
    args.kill_time = 0  # keep track of when first kill signal arrived

    # perform https setup
    if args.use_https:
        https_setup(args, get_pilot_version())
    args.amq = None

    # let the server know that the worker has started
    if args.update_server:
        send_worker_status('started', args.queue, args.url, args.port, logger, 'IPv6')  # note: assuming IPv6, fallback in place

    # initialize InfoService
    try:
        infosys.init(args.queue)
        # check if queue is ACTIVE
        if infosys.queuedata.state != 'ACTIVE':
            logger.critical('specified queue is NOT ACTIVE: %s -- aborting', infosys.queuedata.name)
            return errors.PANDAQUEUENOTACTIVE
    except PilotException as error:
        logger.fatal(error)
        return error.get_error_code()

    # handle special CRIC variables via params
    # internet protocol versions 'IPv4' or 'IPv6' can be set via CRIC PQ.params.internet_protocol_version
    # (must be defined per PQ if wanted). The pilot default is IPv6
    args.internet_protocol_version = infosys.queuedata.params.get('internet_protocol_version', 'IPv6') if infosys.queuedata.params else 'IPv6'
    environ['PILOT_IP_VERSION'] = args.internet_protocol_version

    # set the site name for rucio
    environ['PILOT_RUCIO_SITENAME'] = os.environ.get('PILOT_RUCIO_SITENAME', None) or infosys.queuedata.site

    # store the site name as set with a pilot option
    environ['PILOT_SITENAME'] = infosys.queuedata.resource  #args.site  # TODO: replace with singleton

    # set requested workflow
    logger.info('pilot arguments: %s', str(args))
    workflow = __import__('pilot.workflow.%s' % args.workflow, globals(), locals(), [args.workflow], 0)

    # execute workflow
    try:
        exitcode = workflow.run(args)
    except Exception as exc:
        logger.fatal('main pilot function caught exception: %s', exc)
        exitcode = None

    # let the server know that the worker has finished
    if args.update_server:
        send_worker_status('finished', args.queue, args.url, args.port, logger, args.internet_protocol_version)

    return exitcode


def str2bool(_var):
    """ Helper function to convert string to bool """

    if isinstance(_var, bool):
        return _var
    if _var.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif _var.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def get_args():
    """
    Return the args from the arg parser.

    :return: args (arg parser object).
    """

    arg_parser = argparse.ArgumentParser()

    # pilot log creation
    arg_parser.add_argument('--no-pilot-log',
                            dest='nopilotlog',
                            action='store_true',
                            default=False,
                            help='Do not write the pilot log to file')

    # pilot work directory
    arg_parser.add_argument('-a', '--workdir',
                            dest='workdir',
                            default="",
                            help='Pilot work directory')

    # debug option to enable more log messages
    arg_parser.add_argument('-d', '--debug',
                            dest='debug',
                            action='store_true',
                            default=False,
                            help='Enable debug mode for logging messages')

    # the choices must match in name the python module in pilot/workflow/
    arg_parser.add_argument('-w', '--workflow',
                            dest='workflow',
                            default='generic',
                            choices=['generic', 'generic_hpc',
                                     'production', 'production_hpc',
                                     'analysis', 'analysis_hpc',
                                     'eventservice_hpc', 'stager', 'payload_stageout'],
                            help='Pilot workflow (default: generic)')

    # graciously stop pilot process after hard limit
    arg_parser.add_argument('-l', '--lifetime',
                            dest='lifetime',
                            default=324000,
                            required=False,
                            type=int,
                            help='Pilot lifetime seconds (default: 324000 s)')

    # set the appropriate site, resource and queue
    arg_parser.add_argument('-q', '--queue',
                            dest='queue',
                            required=True,
                            help='MANDATORY: queue name (e.g., AGLT2_TEST-condor)')
    arg_parser.add_argument('-r', '--resource',
                            dest='resource',
                            required=False,  # From v 2.2.0 the resource name is internally set
                            help='OBSOLETE: resource name (e.g., AGLT2_TEST)')
    arg_parser.add_argument('-s', '--site',
                            dest='site',
                            required=False,  # From v 2.2.1 the site name is internally set
                            help='OBSOLETE: site name (e.g., AGLT2_TEST)')

    # graciously stop pilot process after hard limit
    arg_parser.add_argument('-j', '--joblabel',
                            dest='job_label',
                            default='ptest',
                            help='Job prod/source label (default: ptest)')

    # pilot version tag; PR or RC
    arg_parser.add_argument('-i', '--versiontag',
                            dest='version_tag',
                            default='PR',
                            help='Version tag (default: PR, optional: RC)')

    arg_parser.add_argument('-z', '--noserverupdate',
                            dest='update_server',
                            action='store_false',
                            default=True,
                            help='Disable server updates')

    arg_parser.add_argument('-t', '--noproxyverification',
                            dest='verify_proxy',
                            action='store_false',
                            default=True,
                            help='Disable proxy verification')

    arg_parser.add_argument('-u', '--verifypayloadproxy',
                            dest='verify_payload_proxy',
                            action='store_false',
                            default=True,
                            help='Disable payload proxy verification')

    # graciously stop pilot process after hard limit
    arg_parser.add_argument('-v', '--getjobrequests',
                            dest='getjob_requests',
                            default=2,
                            required=False,
                            type=int,
                            help='Number of getjob requests')

    arg_parser.add_argument('-x', '--getjobfailures',
                            dest='getjob_failures',
                            default=5,
                            required=False,
                            type=int,
                            help='Maximum number of getjob request failures in Harvester mode')

    arg_parser.add_argument('--subscribe-to-msgsvc',
                            dest='subscribe_to_msgsvc',
                            action='store_true',
                            default=False,
                            required=False,
                            help='Ask Pilot to receive job/task info from ActiveMQ')

    # SSL certificates
    arg_parser.add_argument('--cacert',
                            dest='cacert',
                            default=None,
                            help='CA certificate to use with HTTPS calls to server, commonly X509 proxy',
                            metavar='path/to/your/certificate')
    arg_parser.add_argument('--capath',
                            dest='capath',
                            default=None,
                            help='CA certificates path',
                            metavar='path/to/certificates/')

    # Server URLs and ports
    arg_parser.add_argument('--url',
                            dest='url',
                            default='',  # the proper default is stored in default.cfg
                            help='PanDA server URL')
    arg_parser.add_argument('-p', '--port',
                            dest='port',
                            default=25443,
                            help='PanDA server port')
    arg_parser.add_argument('--queuedata-url',
                            dest='queuedata_url',
                            default='',
                            help='Queuedata server URL')
    arg_parser.add_argument('--storagedata-url',
                            dest='storagedata_url',
                            default='',
                            help='URL for downloading DDM end points data')
    arg_parser.add_argument('--rucio-host',
                            dest='rucio_host',
                            default='',
                            help='URL for the Rucio host (optional)')
    arg_parser.add_argument('--redirect-stdout',
                            dest='redirectstdout',
                            default='',
                            help='Redirect all stdout to given file, or /dev/null (optional)')

    # Country group
    arg_parser.add_argument('--country-group',
                            dest='country_group',
                            default='',
                            help='Country group option for getjob request')

    # Working group
    arg_parser.add_argument('--working-group',
                            dest='working_group',
                            default='',
                            help='Working group option for getjob request')

    # Allow other country
    arg_parser.add_argument('--allow-other-country',
                            dest='allow_other_country',
                            type=str2bool,
                            default=False,
                            help='Is the resource allowed to be used outside the privileged group?')

    # Allow same user
    arg_parser.add_argument('--allow-same-user',
                            dest='allow_same_user',
                            type=str2bool,
                            default=True,
                            help='Multi-jobs will only come from same taskID (and thus same user)')

    # Experiment
    arg_parser.add_argument('--pilot-user',
                            dest='pilot_user',
                            default='generic',
                            required=True,
                            help='Pilot user (e.g. name of experiment corresponding to pilot plug-in)')

    # Harvester specific options (if any of the following options are used, args.harvester will be set to True)
    arg_parser.add_argument('--harvester-workdir',
                            dest='harvester_workdir',
                            default='',
                            help='Harvester work directory')
    arg_parser.add_argument('--harvester-datadir',
                            dest='harvester_datadir',
                            default='',
                            help='Harvester data directory')
    arg_parser.add_argument('--harvester-eventstatusdump',
                            dest='harvester_eventstatusdump',
                            default='',
                            help='Harvester event status dump json file containing processing status')
    arg_parser.add_argument('--harvester-workerattributes',
                            dest='harvester_workerattributes',
                            default='',
                            help='Harvester worker attributes json file containing job status')
    arg_parser.add_argument('--harvester-submit-mode',
                            dest='harvester_submitmode',
                            default='PULL',
                            help='Harvester submit mode (PUSH or PULL [default])')
    arg_parser.add_argument('--resource-type',
                            dest='resource_type',
                            default='',
                            type=str,
                            choices=['SCORE', 'MCORE', 'SCORE_HIMEM', 'MCORE_HIMEM'],
                            help='Resource type; MCORE, SCORE, SCORE_HIMEM or MCORE_HIMEM')
    arg_parser.add_argument('--use-https',
                            dest='use_https',
                            type=str2bool,
                            default=True,
                            help='Use HTTPS protocol for communications with server')
    arg_parser.add_argument('--cleanup',
                            dest='cleanup',
                            type=str2bool,
                            default=True,
                            help='Cleanup work directory after pilot has finished')
    arg_parser.add_argument('--use-realtime-logging',
                            dest='use_realtime_logging',
                            action='store_true',
                            default=False,
                            help='Use near real-time logging, default=False')
    arg_parser.add_argument('--realtime-logging-server',
                            dest='realtime_logging_server',
                            default=None,
                            help='Realtime logging server [type:host:port]; e.g. logstash:12.23.34.45:80')
    # arg_parser.add_argument('--realtime-logfiles',
    #                         dest='realtime_logfiles',
    #                         default=None,
    #                         help='The log filenames to be sent to the realtime logging server')
    arg_parser.add_argument('--realtime-logname',
                            dest='realtime_logname',
                            default=None,
                            help='The log name in the realtime logging server')

    # Harvester and Nordugrid specific options
    arg_parser.add_argument('--input-dir',
                            dest='input_dir',
                            default='',
                            help='Input directory')
    arg_parser.add_argument('--input-destination-dir',
                            dest='input_destination_dir',
                            default='',
                            help='Input destination directory')
    arg_parser.add_argument('--output-dir',
                            dest='output_dir',
                            default='',
                            help='Output directory')
    arg_parser.add_argument('--job-type',
                            dest='jobtype',
                            default='',
                            help='Job type (managed, user)')
    arg_parser.add_argument('--use-rucio-traces',
                            dest='use_rucio_traces',
                            type=str2bool,
                            default=True,
                            help='Use rucio traces')

    # HPC options
    arg_parser.add_argument('--hpc-resource',
                            dest='hpc_resource',
                            default='',
                            help='Name of the HPC (e.g. Titan)')
    arg_parser.add_argument('--hpc-mode',
                            dest='hpc_mode',
                            default='manytoone',
                            help='HPC mode (manytoone, jumbojobs)')
    arg_parser.add_argument('--es-executor-type',
                            dest='executor_type',
                            default='generic',
                            help='Event service executor type (generic, raythena)')

    return arg_parser.parse_args()


def create_main_work_dir():
    """
    Create and return the pilot's main work directory.
    The function also sets args.mainworkdir and cd's into this directory.
    Note: args, used in this function, is defined in outer scope.

    :return: exit code (int), main work directory (string).
    """

    exitcode = 0

    if args.workdir != "":
        _mainworkdir = get_pilot_work_dir(args.workdir)
        try:
            # create the main PanDA Pilot work directory
            mkdirs(_mainworkdir)
        except PilotException as error:
            # print to stderr since logging has not been established yet
            print('failed to create workdir at %s -- aborting: %s' % (_mainworkdir, error), file=sys.stderr)
            exitcode = shell_exit_code(error._errorCode)
    else:
        _mainworkdir = getcwd()

    args.mainworkdir = _mainworkdir
    chdir(_mainworkdir)

    return exitcode, _mainworkdir


def set_environment_variables():
    """
    Set environment variables. To be replaced with singleton implementation.
    This function sets PILOT_WORK_DIR, PILOT_HOME, PILOT_SITENAME, PILOT_USER and PILOT_VERSION and others.
    Note: args and mainworkdir, used in this function, are defined in outer scope.

    :return:
    """

    # working directory as set with a pilot option (e.g. ..)
    environ['PILOT_WORK_DIR'] = args.workdir  # TODO: replace with singleton

    # main work directory (e.g. /scratch/PanDA_Pilot3_3908_1537173670)
    environ['PILOT_HOME'] = mainworkdir  # TODO: replace with singleton

    # pilot source directory (e.g. /cluster/home/usatlas1/gram_scratch_hHq4Ns/condorg_oqmHdWxz)
    environ['PILOT_SOURCE_DIR'] = args.sourcedir  # TODO: replace with singleton

    # set the pilot user (e.g. ATLAS)
    environ['PILOT_USER'] = args.pilot_user  # TODO: replace with singleton

    # internal pilot state
    environ['PILOT_JOB_STATE'] = 'startup'  # TODO: replace with singleton

    # set the pilot version
    environ['PILOT_VERSION'] = get_pilot_version()

    # set the default wrap-up/finish instruction
    environ['PILOT_WRAP_UP'] = 'NORMAL'

    # proxy verifications
    environ['PILOT_PROXY_VERIFICATION'] = '%s' % args.verify_proxy
    environ['PILOT_PAYLOAD_PROXY_VERIFICATION'] = '%s' % args.verify_payload_proxy

    # keep track of the server updates, if any
    environ['SERVER_UPDATE'] = SERVER_UPDATE_NOT_DONE

    # set the (HPC) resource name (if set in options)
    environ['PILOT_RESOURCE_NAME'] = args.hpc_resource

    # allow for the possibility of turning off rucio traces
    environ['PILOT_USE_RUCIO_TRACES'] = str(args.use_rucio_traces)

    # event service executor type
    environ['PILOT_ES_EXECUTOR_TYPE'] = args.executor_type

    if args.output_dir:
        environ['PILOT_OUTPUT_DIR'] = args.output_dir

    # keep track of the server urls
    environ['PANDA_SERVER_URL'] = get_panda_server(args.url, args.port, update_server=args.update_server)
    environ['QUEUEDATA_SERVER_URL'] = '%s' % args.queuedata_url
    if args.storagedata_url:
        environ['STORAGEDATA_SERVER_URL'] = '%s' % args.storagedata_url


def wrap_up():
    """
    Perform cleanup and terminate logging.
    Note: args and mainworkdir, used in this function, are defined in outer scope.

    :return: exit code (int).
    """

    exitcode = 0

    # cleanup pilot workdir if created
    if args.sourcedir != mainworkdir and args.cleanup:
        chdir(args.sourcedir)
        try:
            rmtree(mainworkdir)
        except Exception as exc:
            logging.warning("failed to remove %s: %s", mainworkdir, exc)
        else:
            logging.info("removed %s", mainworkdir)

    # in Harvester mode, create a kill_worker file that will instruct Harvester that the pilot has finished
    if args.harvester:
        from pilot.util.harvester import kill_worker
        kill_worker()

    try:
        exitcode = trace.pilot['error_code']
    except Exception:
        exitcode = trace
    else:
        logging.info('traces error code: %d', exitcode)
        if trace.pilot['nr_jobs'] <= 1:
            if exitcode != 0:
                logging.info('an exit code was already set: %d (will be converted to a standard shell code)', exitcode)
        elif trace.pilot['nr_jobs'] > 0:
            if trace.pilot['nr_jobs'] == 1:
                logging.getLogger(__name__).info('pilot has finished (%d job was processed)', trace.pilot['nr_jobs'])
            else:
                logging.getLogger(__name__).info('pilot has finished (%d jobs were processed)', trace.pilot['nr_jobs'])
            exitcode = SUCCESS
        elif trace.pilot['state'] == FAILURE:
            logging.critical('pilot workflow failure -- aborting')
        elif trace.pilot['state'] == ERRNO_NOJOBS:
            logging.critical('pilot did not process any events -- aborting')
            exitcode = ERRNO_NOJOBS

    try:
        exitcode = int(exitcode)
    except TypeError as exc:
        logging.warning(f'failed to convert exit code to int: {exitcode}, {exc}')
        exitcode = 1008

    logging.info('pilot has finished')
    logging.shutdown()

    return shell_exit_code(exitcode)


def get_pilot_source_dir():
    """
    Return the pilot source directory.

    :return: full path to pilot source directory.
    """

    cwd = getcwd()
    if exists(join(join(cwd, 'pilot3'), 'pilot.py')):  # in case wrapper has untarred src as pilot3 in init dir
        return join(cwd, 'pilot3')
    elif exists(join(cwd, 'pilot.py')):  # in case pilot gets launched from within the src dir
        return cwd
    else:
        # could throw error here, but logging is not setup yet - fail later
        return cwd


def send_worker_status(status, queue, url, port, logger, internet_protocol_version):
    """
    Send worker info to the server to let it know that the worker has started
    Note: the function can fail, but if it does, it will be ignored.

    :param status: 'started' or 'finished' (string).
    :param queue: PanDA queue name (string).
    :param url: server url (string).
    :param port: server port (string).
    :param logger: logging object.
    :param internet_protocol_version: internet protocol version, IPv4 or IPv6 (string).
    :return:
    """

    # worker node structure to be sent to the server
    data = {}
    data['workerID'] = os.environ.get('HARVESTER_WORKER_ID', None)
    data['harvesterID'] = os.environ.get('HARVESTER_ID', None)
    data['status'] = status
    data['site'] = queue

    # attempt to send the worker info to the server
    if data['workerID'] and data['harvesterID']:
        send_update('updateWorkerPilotStatus', data, url, port, ipv=internet_protocol_version)
    else:
        logger.warning('workerID/harvesterID not known, will not send worker status to server')


def set_lifetime(args):
    """
    Update the pilot lifetime if set by an environment variable (PANDAPILOT_LIFETIME) (in seconds).

    :param args: pilot args.
    :return:
    """

    lifetime = os.environ.get('PANDAPILOT_LIFETIME', None)
    if lifetime:
        try:
            lifetime = int(lifetime)
        except (ValueError, TypeError):
            pass
        else:
            args.lifetime = lifetime


def set_redirectall(args):
    """

    """

    redirectall = os.environ.get('PANDAPILOT_REDIRECTALL', False)
    if not redirectall:
        try:
            redirectall = bool(redirectall)
        except (ValueError, TypeError):
            pass
        else:
            args.redirectall = redirectall


if __name__ == '__main__':
    """
    Main function of pilot module.
    """

    # get the args from the arg parser
    args = get_args()
    args.last_heartbeat = time.time()
    # args.rucio_host = 'https://voatlasrucio-server-prod.cern.ch:443'

    # Define and set the main harvester control boolean
    args.harvester = is_harvester_mode(args)

    # initialize the pilot timing dictionary
    args.timing = {}  # TODO: move to singleton?

    # initialize job status dictionary (e.g. used to keep track of log transfers)
    args.job_status = {}  # TODO: move to singleton or to job object directly?

    #if 'BNL_OSG_SPHENIX_TEST' in args.queue:
    #    args.lifetime = 3600
    #    args.subscribe_to_msgsvc = True
    #    args.redirectstdout = '/dev/null'

    # store T0 time stamp
    add_to_pilot_timing('0', PILOT_START_TIME, time.time(), args)
    add_to_pilot_timing('1', PILOT_MULTIJOB_START_TIME, time.time(), args)

    # if requested by the wrapper via a pilot option, create the main pilot workdir and cd into it
    args.sourcedir = getcwd()  #get_pilot_source_dir()

    exit_code, mainworkdir = create_main_work_dir()
    if exit_code != 0:
        sys.exit(exit_code)

    set_lifetime(args)

    # setup and establish standard logging
    establish_logging(debug=args.debug, nopilotlog=args.nopilotlog, redirectstdout=args.redirectstdout)

    # set environment variables (to be replaced with singleton implementation)
    set_environment_variables()

    # execute main function
    trace = main()

    # store final time stamp (cannot be placed later since the mainworkdir is about to be purged)
    add_to_pilot_timing('0', PILOT_END_TIME, time.time(), args, store=False)

    # perform cleanup and terminate logging
    exit_code = wrap_up()

    # the end.
    sys.exit(exit_code)



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Danila Oleynik, danila.oleynik@cern.ch, 2018
# - Paul Nilsson, paul.nilsson@cern.ch, 2019-2022

# Note: The Pilot utilities to provide MPI related functionality through mpi4py
# Required for HPC workflow where the Pilot acts like an MPI application

# remove logging for now since it has a tendency to dump error messages like this in stderr:
# 'No handlers could be found for logger "pilot.util.mpi"'
# import logging
# try:
#     logger = logging.getLogger(__name__)
# except Exception:
#     logger = None


def get_ranks_info():
    """
    Return current MPI rank and number of ranks
    None, None - if MPI environment is not available

    :return: rank, max_rank
    """

    rank = None
    max_rank = None
    try:
        pass  # removed in April 2020 since it was only used on Titan and is causing import errors
        #from mpi4py import MPI
        #comm = MPI.COMM_WORLD
        #rank = comm.Get_rank()
        #max_rank = comm.Get_size()
    except ImportError:
        print("mpi4py not found")
    return rank, max_rank



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2017-2022

import os
from re import findall
from signal import SIGTERM

from pilot.common.exception import TrfDownloadFailure
from pilot.util.config import config
from pilot.util.constants import UTILITY_BEFORE_PAYLOAD, UTILITY_AFTER_PAYLOAD_STARTED
from pilot.util.filehandling import read_file
from .setup import get_analysis_trf

import logging
logger = logging.getLogger(__name__)


def sanity_check():
    """
    Perform an initial sanity check before doing anything else in a given workflow.
    This function can be used to verify importing of modules that are otherwise used much later, but it is better to abort
    the pilot if a problem is discovered early.

    :return: exit code (0 if all is ok, otherwise non-zero exit code).
    """

    return 0


def validate(job):
    """
    Perform user specific payload/job validation.

    :param job: job object.
    :return: Boolean (True if validation is successful).
    """

    return True


def get_payload_command(job):
    """
    Return the full command for executing the payload, including the sourcing of all setup files and setting of
    environment variables.

    By default, the full payload command is assumed to be in the job.jobparams.

    :param job: job object
    :return: command (string)
    """

    # Try to download the trf
    # if job.imagename != "" or "--containerImage" in job.jobparams:
    #    job.transformation = os.path.join(os.path.dirname(job.transformation), "runcontainer")
    #    logger.warning('overwrote job.transformation, now set to: %s' % job.transformation)
    ec, diagnostics, trf_name = get_analysis_trf(job.transformation, job.workdir)
    if ec != 0:
        raise TrfDownloadFailure(diagnostics)
    else:
        logger.debug('user analysis trf: %s' % trf_name)

    return get_analysis_run_command(job, trf_name)


def get_analysis_run_command(job, trf_name):
    """
    Return the proper run command for the user job.

    Example output: export X509_USER_PROXY=<..>;./runAthena <job parameters> --usePFCTurl --directIn

    :param job: job object.
    :param trf_name: name of the transform that will run the job (string). Used when containers are not used.
    :return: command (string).
    """

    cmd = ""

    # add the user proxy
    if 'X509_USER_PROXY' in os.environ and not job.imagename:
        cmd += 'export X509_USER_PROXY=%s;' % os.environ.get('X509_USER_PROXY')

    # set up trfs
    if job.imagename == "":  # user jobs with no imagename defined
        cmd += './%s %s' % (trf_name, job.jobparams)
    else:
        if trf_name:
            cmd += './%s %s' % (trf_name, job.jobparams)
        else:
            cmd += 'python %s %s' % (trf_name, job.jobparams)

    return cmd


def update_job_data(job):
    """
    This function can be used to update/add data to the job object.
    E.g. user specific information can be extracted from other job object fields. In the case of ATLAS, information
    is extracted from the metaData field and added to other job object fields.

    :param job: job object
    :return:
    """

    pass


def remove_redundant_files(workdir, outputfiles=None, piloterrors=[], debugmode=False):
    """
    Remove redundant files and directories prior to creating the log file.

    :param workdir: working directory (string).
    :param outputfiles: list of output files.
    :param piloterrors: list of Pilot assigned error codes (list).
    :param debugmode: True if debug mode has been switched on (Boolean).
    :return:
    """

    pass


def get_utility_commands(order=None, job=None):
    """
    Return a dictionary of utility commands and arguments to be executed in parallel with the payload.
    This could e.g. be memory and network monitor commands. A separate function can be used to determine the
    corresponding command setups using the utility command name.
    If the optional order parameter is set, the function should return the list of corresponding commands.
    E.g. if order=UTILITY_BEFORE_PAYLOAD, the function should return all commands that are to be executed before the
    payload. If order=UTILITY_WITH_PAYLOAD, the corresponding commands will be prepended to the payload execution
    string. If order=UTILITY_AFTER_PAYLOAD_STARTED, the commands that should be executed after the payload has been started
    should be returned.

    FORMAT: {'command': <command>, 'args': <args>}

    :param order: optional sorting order (see pilot.util.constants)
    :param job: optional job object.
    :return: dictionary of utilities to be executed in parallel with the payload.
    """

    return {}


def get_utility_command_setup(name, job, setup=None):
    """
    Return the proper setup for the given utility command.
    If a payload setup is specified
    :param name:
    :param setup:
    :return:
    """

    pass


def get_utility_command_execution_order(name):
    """
    Should the given utility command be executed before or after the payload?

    :param name: utility name (string).
    :return: execution order constant (UTILITY_BEFORE_PAYLOAD or UTILITY_AFTER_PAYLOAD_STARTED)
    """

    # example implementation
    if name == 'monitor':
        return UTILITY_BEFORE_PAYLOAD
    else:
        return UTILITY_AFTER_PAYLOAD_STARTED


def post_utility_command_action(name, job):
    """
    Perform post action for given utility command.

    :param name: name of utility command (string).
    :param job: job object.
    :return:
    """

    pass


def get_utility_command_kill_signal(name):
    """
    Return the proper kill signal used to stop the utility command.

    :param name:
    :return: kill signal
    """

    return SIGTERM


def get_utility_command_output_filename(name, selector=None):
    """
    Return the filename to the output of the utility command.

    :param name: utility name (string).
    :param selector: optional special conditions flag (boolean).
    :return: filename (string).
    """

    return ""


def verify_job(job):
    """
    Verify job parameters for specific errors.
    Note:
      in case of problem, the function should set the corresponding pilot error code using
      job.piloterrorcodes, job.piloterrordiags = errors.add_error_code(error.get_error_code())

    :param job: job object
    :return: Boolean.
    """

    return True


def update_stagein(job):
    """
    In case special files need to be skipped during stage-in, the job.indata list can be updated here.
    See ATLAS code for an example.

    :param job: job object.
    :return:
    """

    pass


def get_metadata(workdir):
    """
    Return the metadata from file.

    :param workdir: work directory (string)
    :return:
    """

    path = os.path.join(workdir, config.Payload.jobreport)
    metadata = read_file(path) if os.path.exists(path) else None

    return metadata


def update_server(job):
    """
    Perform any user specific server actions.

    E.g. this can be used to send special information to a logstash.

    :param job: job object.
    :return:
    """

    pass


def post_prestagein_utility_command(**kwargs):
    """
    Execute any post pre-stage-in utility commands.

    :param kwargs: kwargs (dictionary).
    :return:
    """

    # label = kwargs.get('label', 'unknown_label')
    # stdout = kwargs.get('output', None)

    pass


def process_debug_command(debug_command, pandaid):
    """
    In debug mode, the server can send a special debug command to the pilot via the updateJob backchannel.
    This function can be used to process that command, i.e. to identify a proper pid to debug (which is unknown
    to the server).

    :param debug_command: debug command (string), payload pid (int).
    :param pandaid: PanDA id (string).
    :return: updated debug command (string)
    """

    return debug_command


def allow_timefloor(submitmode):
    """
    Should the timefloor mechanism (multi-jobs) be allowed for the given submit mode?

    :param submitmode: submit mode (string).
    """

    allow = True
    if submitmode.lower() == 'push':
        logger.info('Since the submitmode=push, override timefloor with zero manually')
        allow = False

    return allow


def get_pilot_id(jobid):
    """
    Get the pilot id from the environment variable GTAG.
    Update for each job to get a unique pilot id per job.

    :param jobid: PanDA job id (int).
    :return: pilot id (string).
    """

    pilotid = os.environ.get("GTAG", "unknown")
    regex = r'PandaJob\_(\d+)+'
    _id = findall(regex, pilotid)
    if _id:
        pilotid = pilotid.replace(_id[0], str(jobid))

    return pilotid



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2021

# based on Harvester k8s_utils

import os

from kubernetes import client, config
# from kubernetes.client.rest import ApiException


class K8sClient(object):

    def __init__(self, namespace, config_file=None):
        """
        Init.
        """

        if not os.path.isfile(config_file):
            raise RuntimeError('Cannot find k8s config file: {0}'.format(config_file))

        config.load_kube_config(config_file=config_file)
        self.namespace = namespace if namespace else 'default'
        self.core_v1 = client.CoreV1Api()
        self.batch_v1 = client.BatchV1Api()
        self.apps_v1 = client.AppsV1Api()
        self.delete_v1 = client.V1DeleteOptions(propagation_policy='Background')

    def create_ssh_keys_secret(self, work_spec):
        return True

    def create_dask_head(self, work_spec, panda_queue, evaluation_image, pilot_image, cert, dfs_claim_name,
                         cpu_adjust_ratio=100, memory_adjust_ratio=100, max_time=None):
        """

        """

        #worker_id = str(work_spec.workerID)
        #tmp_log = core_utils.make_logger(base_logger, 'workerID={0}'.format(worker_id),
        #                                 method_name='create_dask_head')

        if not self.create_configmap_dask(work_spec, panda_queue):
            return False

        if not self.create_host_discovery_configmap(work_spec):
            return False

    def create_dask_formation(self, work_spec, panda_queue, evaluation_image, pilot_image,
                              worker_command, cert, dfs_claim_name, cpu_adjust_ratio=100, memory_adjust_ratio=100,
                              max_time=None):

        rsp = self.create_ssh_keys_secret(work_spec)
        if not rsp:
            return rsp

        rsp = self.create_dask_head(work_spec, panda_queue, evaluation_image, pilot_image, cert, dfs_claim_name,
                                    cpu_adjust_ratio, memory_adjust_ratio, max_time)
        if not rsp:
            return rsp

        rsp = self.create_dask_workers(work_spec, evaluation_image, worker_command, dfs_claim_name,
                                       cpu_adjust_ratio, memory_adjust_ratio, max_time)

        if not rsp:
            return rsp

        return True



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Pavlo Svirin, pavlo.svirin@cern.ch, 2017
# - Tobias Wegner, tobias.wegner@cern.ch, 2018
# - Paul Nilsson, paul.nilsson@cern.ch, 2018-2021

import os
import logging
import errno
from time import time

from .common import get_copysetup, verify_catalog_checksum, resolve_common_transfer_errors  #, get_timeout
from pilot.common.exception import StageInFailure, StageOutFailure, PilotException, ErrorCodes
from pilot.util.container import execute
#from pilot.util.timer import timeout


logger = logging.getLogger(__name__)

require_replicas = True  ## indicate if given copytool requires input replicas to be resolved

allowed_schemas = ['srm', 'gsiftp', 'root']  # prioritized list of supported schemas for transfers by given copytool


def is_valid_for_copy_in(files):
    return True  ## FIX ME LATER
    #for f in files:
    #    if not all(key in f for key in ('name', 'source', 'destination')):
    #        return False
    #return True


def is_valid_for_copy_out(files):
    #for f in files:
    #    if not all(key in f for key in ('name', 'source', 'destination')):
    #        return False
    return True


def copy_in_old(files):
    """
    Tries to download the given files using lsm-get directly.

    :param files: Files to download
    :raises PilotException: StageInFailure
    """

    if not check_for_lsm(dst_in=True):
        raise StageInFailure("No LSM tools found")
    exit_code, stdout, stderr = move_all_files_in(files)
    if exit_code != 0:
        # raise failure
        raise StageInFailure(stdout)


def copy_in(files, **kwargs):
    """
    Download given files using the lsm-get command.

    :param files: list of `FileSpec` objects.
    :raise: PilotException in case of controlled error.
    :return: files `FileSpec` object.
    """

    exit_code = 0
    stdout = ""
    stderr = ""

    copytools = kwargs.get('copytools') or []
    copysetup = get_copysetup(copytools, 'lsm')
    trace_report = kwargs.get('trace_report')
    #allow_direct_access = kwargs.get('allow_direct_access')

    # note, env vars might be unknown inside middleware contrainers, if so get the value already in the trace report
    localsite = os.environ.get('RUCIO_LOCAL_SITE_ID', trace_report.get_value('localSite'))

    for fspec in files:
        # update the trace report
        localsite = localsite if localsite else fspec.ddmendpoint
        trace_report.update(localSite=localsite, remoteSite=fspec.ddmendpoint, filesize=fspec.filesize)
        trace_report.update(filename=fspec.lfn, guid=fspec.guid.replace('-', ''))
        trace_report.update(scope=fspec.scope, dataset=fspec.dataset)

        # continue loop for files that are to be accessed directly  ## TO BE DEPRECATED (anisyonk)
        #if fspec.is_directaccess(ensure_replica=False) and allow_direct_access and fspec.accessmode == 'direct':
        #    fspec.status_code = 0
        #    fspec.status = 'remote_io'
        #    trace_report.update(url=fspec.turl, clientState='FOUND_ROOT', stateReason='direct_access')
        #    trace_report.send()
        #    continue

        trace_report.update(catStart=time())

        dst = fspec.workdir or kwargs.get('workdir') or '.'
        #timeout = get_timeout(fspec.filesize)
        source = fspec.turl
        destination = os.path.join(dst, fspec.lfn)

        logger.info("transferring file %s from %s to %s", fspec.lfn, source, destination)

        exit_code, stdout, stderr = move(source, destination, dst_in=True, copysetup=copysetup)

        if exit_code != 0:
            logger.warning("transfer failed: exit code = %d, stdout = %s, stderr = %s", exit_code, stdout, stderr)

            error = resolve_common_transfer_errors(stderr, is_stagein=True)
            fspec.status = 'failed'
            fspec.status_code = error.get('rcode')
            trace_report.update(clientState=error.get('state') or 'STAGEIN_ATTEMPT_FAILED',
                                stateReason=error.get('error'), timeEnd=time())
            trace_report.send()
            raise PilotException(error.get('error'), code=error.get('rcode'), state=error.get('state'))

        # verify checksum; compare local checksum with catalog value (fspec.checksum), use same checksum type
        state, diagnostics = verify_catalog_checksum(fspec, destination)
        if diagnostics != "":
            trace_report.update(clientState=state or 'STAGEIN_ATTEMPT_FAILED', stateReason=diagnostics,
                                timeEnd=time())
            trace_report.send()
            raise PilotException(diagnostics, code=fspec.status_code, state=state)

        fspec.status_code = 0
        fspec.status = 'transferred'
        trace_report.update(clientState='DONE', stateReason='OK', timeEnd=time())
        trace_report.send()

    # for testing kill signals
    #import signal
    #os.kill(os.getpid(), signal.SIGSEGV)

    return files


def copy_out(files, **kwargs):
    """
    Upload given files using lsm copytool.

    :param files: list of `FileSpec` objects.
    :raise: PilotException in case of controlled error.
    """

    copytools = kwargs.get('copytools') or []
    copysetup = get_copysetup(copytools, 'lsm')
    trace_report = kwargs.get('trace_report')
    ddmconf = kwargs.get('ddmconf', None)
    if not ddmconf:
        raise PilotException("copy_out() failed to resolve ddmconf from function arguments",
                             code=ErrorCodes.STAGEOUTFAILED,
                             state='COPY_ERROR')

    for fspec in files:
        trace_report.update(scope=fspec.scope, dataset=fspec.dataset, url=fspec.surl, filesize=fspec.filesize)
        trace_report.update(catStart=time(), filename=fspec.lfn, guid=fspec.guid.replace('-', ''))

        # resolve token value from fspec.ddmendpoint
        ddm = ddmconf.get(fspec.ddmendpoint)
        token = ddm.token
        if not token:
            diagnostics = "copy_out() failed to resolve token value for ddmendpoint=%s" % (fspec.ddmendpoint)
            trace_report.update(clientState='STAGEOUT_ATTEMPT_FAILED',
                                stateReason=diagnostics,
                                timeEnd=time())
            trace_report.send()
            raise PilotException(diagnostics, code=ErrorCodes.STAGEOUTFAILED, state='COPY_ERROR')

        src = fspec.workdir or kwargs.get('workdir') or '.'
        #timeout = get_timeout(fspec.filesize)
        source = os.path.join(src, fspec.lfn)
        destination = fspec.turl

        # checksum has been calculated in the previous step - transfer_files() in api/data
        # note: pilot is handing over checksum to the command - which will/should verify it after the transfer
        checksum = "adler32:%s" % fspec.checksum.get('adler32')

        # define the command options
        opts = {'--size': fspec.filesize,
                '-t': token,
                '--checksum': checksum,
                '--guid': fspec.guid}
        try:
            opts = " ".join(["%s %s" % (k, v) for (k, v) in opts.iteritems()])  # Python 2
        except Exception:
            opts = " ".join(["%s %s" % (k, v) for (k, v) in list(opts.items())])  # Python 3

        logger.info("transferring file %s from %s to %s", fspec.lfn, source, destination)

        nretries = 1  # input parameter to function?
        for retry in range(nretries):
            exit_code, stdout, stderr = move(source, destination, dst_in=False, copysetup=copysetup, options=opts)

            if exit_code != 0:
                if stderr == "":
                    stderr = stdout
                error = resolve_common_transfer_errors(stderr, is_stagein=False)
                fspec.status = 'failed'
                fspec.status_code = error.get('exit_code')
                trace_report.update(clientState=error.get('state', None) or 'STAGEOUT_ATTEMPT_FAILED',
                                    stateReason=error.get('error', 'unknown error'),
                                    timeEnd=time())
                trace_report.send()
                raise PilotException(error.get('error'), code=error.get('exit_code'), state=error.get('state'))
            else:  # all successful
                logger.info('all successful')
                break

        fspec.status_code = 0
        fspec.status = 'transferred'
        trace_report.update(clientState='DONE', stateReason='OK', timeEnd=time())
        trace_report.send()

    return files


def copy_out_old(files):
    """
    Tries to upload the given files using lsm-put directly.

    :param files: Files to upload
    :raises PilotException: StageOutFailure
    """

    if not check_for_lsm(dst_in=False):
        raise StageOutFailure("No LSM tools found")

    exit_code, stdout, stderr = move_all_files_out(files)
    if exit_code != 0:
        # raise failure
        raise StageOutFailure(stdout)


def move_all_files_in(files, nretries=1):
    """
    Move all files.

    :param files:
    :param nretries: number of retries; sometimes there can be a timeout copying, but the next attempt may succeed
    :return: exit_code, stdout, stderr
    """

    exit_code = 0
    stdout = ""
    stderr = ""

    for entry in files:  # entry = {'name':<filename>, 'source':<dir>, 'destination':<dir>}
        logger.info("transferring file %s from %s to %s", entry['name'], entry['source'], entry['destination'])

        source = entry['source'] + '/' + entry['name']
        destination = os.path.join(entry['destination'], entry['name'])
        for retry in range(nretries):
            exit_code, stdout, stderr = move(source, destination, dst_in=True)

            if exit_code != 0:
                if ((exit_code != errno.ETIMEDOUT) and (exit_code != errno.ETIME)) or (retry + 1) == nretries:
                    logger.warning("transfer failed: exit code = %d, stdout = %s, stderr = %s", exit_code, stdout, stderr)
                    return exit_code, stdout, stderr
            else:  # all successful
                break

    return exit_code, stdout, stderr


def move_all_files_out(files, nretries=1):
    """
    Move all files.

    :param files:
    :return: exit_code, stdout, stderr
    """

    exit_code = 0
    stdout = ""
    stderr = ""

    for entry in files:  # entry = {'name':<filename>, 'source':<dir>, 'destination':<dir>}
        logger.info("transferring file %s from %s to %s", entry['name'], entry['source'], entry['destination'])

        destination = entry['destination'] + '/' + entry['name']
        source = os.path.join(entry['source'], entry['name'])
        for retry in range(nretries):
            exit_code, stdout, stderr = move(source, destination, dst_in=False)

            if exit_code != 0:
                if ((exit_code != errno.ETIMEDOUT) and (exit_code != errno.ETIME)) or (retry + 1) == nretries:
                    logger.warning("transfer failed: exit code = %d, stdout = %s, stderr = %s", exit_code, stdout, stderr)
                    return exit_code, stdout, stderr
            else:  # all successful
                break

    return exit_code, stdout, stderr


#@timeout(seconds=10800)
def move(source, destination, dst_in=True, copysetup="", options=None):
    """
    Use lsm-get or lsm-put to transfer the file.

    :param source: path to source (string).
    :param destination: path to destination (string).
    :param dst_in: True for stage-in, False for stage-out (boolean).
    :return: exit code, stdout, stderr
    """

    # copysetup = '/osg/mwt2/app/atlas_app/atlaswn/setup.sh'
    if copysetup != "":
        cmd = 'source %s;' % copysetup
    else:
        cmd = ''

    args = "%s %s" % (source, destination)
    if options:
        args = "%s %s" % (options, args)

    if dst_in:
        cmd += "lsm-get %s" % args
    else:
        cmd += "lsm-put %s" % args

    try:
        exit_code, stdout, stderr = execute(cmd, usecontainer=False, copytool=True)  #, timeout=get_timeout(fspec.filesize))
    except Exception as error:
        if dst_in:
            exit_code = ErrorCodes.STAGEINFAILED
        else:
            exit_code = ErrorCodes.STAGEOUTFAILED
        stdout = 'exception caught: e' % error
        stderr = ''
        logger.warning(stdout)

    logger.info('exit_code=%d, stdout=%s, stderr=%s', exit_code, stdout, stderr)
    return exit_code, stdout, stderr


def check_for_lsm(dst_in=True):
    cmd = None
    if dst_in:
        cmd = 'which lsm-get'
    else:
        cmd = 'which lsm-put'
    exit_code, gfal_path, _ = execute(cmd)
    return exit_code == 0
